from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeVar

from .fsharp_collections import ComparisonIdentity_Structural, HashIdentity_Structural
from .system_text import StringBuilder, StringBuilder__Append_Z721C83C5
from .util import IComparer_1, IEqualityComparer, IEqualityComparer_1, dispose, equals, ignore, structural_hash


_T = TypeVar("_T")

_R = TypeVar("_R")

__B = TypeVar("__B")

__A = TypeVar("__A")


class ObjectExpr5(IEqualityComparer):
    def System_Collections_IEqualityComparer_Equals541DA560(self, x: Any = None, y: Any = None) -> bool:
        return equals(x, y)

    def System_Collections_IEqualityComparer_GetHashCode4E60E31B(self, x_1: Any = None) -> int:
        return structural_hash(x_1)


LanguagePrimitives_GenericEqualityComparer: IEqualityComparer = ObjectExpr5()


class ObjectExpr6(IEqualityComparer):
    def System_Collections_IEqualityComparer_Equals541DA560(self, x: Any = None, y: Any = None) -> bool:
        return equals(x, y)

    def System_Collections_IEqualityComparer_GetHashCode4E60E31B(self, x_1: Any = None) -> int:
        return structural_hash(x_1)


LanguagePrimitives_GenericEqualityERComparer: IEqualityComparer = ObjectExpr6()


def LanguagePrimitives_FastGenericComparer(__unit: None = None) -> IComparer_1[Any]:
    return ComparisonIdentity_Structural()


def LanguagePrimitives_FastGenericComparerFromTable(__unit: None = None) -> IComparer_1[Any]:
    return ComparisonIdentity_Structural()


def LanguagePrimitives_FastGenericEqualityComparer(__unit: None = None) -> IEqualityComparer_1[Any]:
    return HashIdentity_Structural()


def LanguagePrimitives_FastGenericEqualityComparerFromTable(__unit: None = None) -> IEqualityComparer_1[Any]:
    return HashIdentity_Structural()


def Operators_Failure(message: str) -> Exception:
    return Exception(message)


def Operators_FailurePattern(exn: Exception) -> str | None:
    return str(exn)


def Operators_NullArg(x: str) -> Any:
    raise Exception(x)


def Operators_Using(resource: Any, action: Callable[[_T], _R]) -> Any:
    try:
        return action(resource)

    finally:
        if equals(resource, None):
            pass

        else:
            copy_of_struct: _T = resource
            dispose(copy_of_struct)


def Operators_Lock(_lockObj: Any, action: Callable[[], __B]) -> Any:
    return action(None)


def ExtraTopLevelOperators_LazyPattern(input: Any) -> Any:
    return input.Value


def PrintfModule_PrintFormatToStringBuilderThen(
    continuation: Callable[[], __A], builder: StringBuilder, format: Any
) -> Any:
    def append(s: str, continuation: Any = continuation, builder: Any = builder, format: Any = format) -> __A:
        ignore(StringBuilder__Append_Z721C83C5(builder, s))
        return continuation(None)

    return format.cont(append)


def PrintfModule_PrintFormatToStringBuilder(builder: StringBuilder, format: Any) -> Any:
    def _arrow13(__unit: None = None, builder: Any = builder, format: Any = format) -> None:
        ignore(None)

    return PrintfModule_PrintFormatToStringBuilderThen(_arrow13, builder, format)


__all__ = [
    "LanguagePrimitives_GenericEqualityComparer",
    "LanguagePrimitives_GenericEqualityERComparer",
    "LanguagePrimitives_FastGenericComparer",
    "LanguagePrimitives_FastGenericComparerFromTable",
    "LanguagePrimitives_FastGenericEqualityComparer",
    "LanguagePrimitives_FastGenericEqualityComparerFromTable",
    "Operators_Failure",
    "Operators_FailurePattern",
    "Operators_NullArg",
    "Operators_Using",
    "Operators_Lock",
    "ExtraTopLevelOperators_LazyPattern",
    "PrintfModule_PrintFormatToStringBuilderThen",
    "PrintfModule_PrintFormatToStringBuilder",
]



from __future__ import annotations
from typing import Any
from ..arctrl_core.arc_types import (ArcStudy, ArcAssay)
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.collections_ import (ResizeArray_choose, ResizeArray_isEmpty)
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.Table.arc_table import ArcTable
from ..fable_library.list import (FSharpList, empty)
from ..fable_library.option import (default_arg, to_array)
from ..fable_library.seq import (iterate_indexed, append, map, try_find, try_pick, iterate)
from ..fable_library.string_ import (to_fail, printf, to_console)
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, to_enumerable, get_enumerator, IEnumerator, ignore)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .AnnotationTable.arc_table import (try_from_fs_worksheet, to_fs_worksheet as to_fs_worksheet_1)
from .DataMapTable.data_map_table import (try_from_fs_worksheet as try_from_fs_worksheet_1, to_fs_worksheet)
from .Metadata.sparse_table import (SparseRowModule_writeToSheet, SparseRowModule_fromValues, SparseRowModule_fromFsRow, SparseRowModule_getAllValues, SparseRowModule_fromAllValues)
from .Metadata.study import (to_rows, from_rows)

def ArcStudy_toMetadataSheet(study: ArcStudy, assays: FSharpList[ArcAssay] | None=None) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet("isa_study")
    def action(row_i: int, r: IEnumerable_1[tuple[int, str]], study: Any=study, assays: Any=assays) -> None:
        SparseRowModule_writeToSheet(row_i + 1, r, sheet)

    def _arrow3639(__unit: None=None, study: Any=study, assays: Any=assays) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        source_1: IEnumerable_1[IEnumerable_1[tuple[int, str]]] = to_rows(study, assays)
        return append(to_enumerable([SparseRowModule_fromValues(to_enumerable(["STUDY"]))]), source_1)

    iterate_indexed(action, _arrow3639())
    return sheet


def ArcStudy_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> tuple[ArcStudy, FSharpList[ArcAssay]] | None:
    en: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(rows)
    ignore(en.System_Collections_IEnumerator_MoveNext())
    return from_rows(2, en)[3]


def ArcStudy_fromMetadataSheet(sheet: FsWorksheet) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
    try: 
        return default_arg(ArcStudy_fromRows(map(SparseRowModule_fromFsRow, sheet.Rows)), (ArcStudy.create(create_missing_identifier()), empty()))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcStudy_toMetadataCollection(study: ArcStudy, assays: FSharpList[ArcAssay] | None=None) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], study: Any=study, assays: Any=assays) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    def _arrow3640(__unit: None=None, study: Any=study, assays: Any=assays) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        source_1: IEnumerable_1[IEnumerable_1[tuple[int, str]]] = to_rows(study, assays)
        return append(to_enumerable([SparseRowModule_fromValues(to_enumerable(["STUDY"]))]), source_1)

    return map(mapping, _arrow3640())


def ArcStudy_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
    try: 
        return default_arg(ArcStudy_fromRows(map(SparseRowModule_fromAllValues, collection)), (ArcStudy.create(create_missing_identifier()), empty()))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcStudy_isMetadataSheetName(name: str) -> bool:
    if name == "isa_study":
        return True

    else: 
        return name == "Study"



def ArcStudy_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return ArcStudy_isMetadataSheetName(sheet.Name)


def ArcStudy_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return ArcStudy_isMetadataSheet(sheet)

    return try_find(predicate, doc.GetWorksheets())


def ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D(doc: FsWorkbook) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
    try: 
        pattern_input: tuple[ArcStudy, FSharpList[ArcAssay]]
        match_value: FsWorksheet | None = ArcStudy_tryGetMetadataSheet(doc)
        if match_value is None:
            to_console(printf("Cannot retrieve metadata: Study file does not contain \"%s\" or \"%s\" sheet."))("isa_study")("Study")
            pattern_input = (ArcStudy.create(create_missing_identifier()), empty())

        else: 
            pattern_input = ArcStudy_fromMetadataSheet(match_value)

        study_metadata: ArcStudy = pattern_input[0]
        sheets: Array[FsWorksheet] = doc.GetWorksheets()
        annotation_tables: Array[ArcTable] = ResizeArray_choose(try_from_fs_worksheet, sheets)
        datamap_sheet: DataMap | None = try_pick(try_from_fs_worksheet_1, sheets)
        if not ResizeArray_isEmpty(annotation_tables):
            study_metadata.Tables = annotation_tables

        study_metadata.DataMap = datamap_sheet
        return (study_metadata, pattern_input[1])

    except Exception as err:
        arg_2: str = str(err)
        return to_fail(printf("Could not parse study: \n%s"))(arg_2)



def ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(study: ArcStudy, assays: FSharpList[ArcAssay] | None=None, datamap_sheet: bool | None=None) -> FsWorkbook:
    datamap_sheet_1: bool = default_arg(datamap_sheet, True)
    doc: FsWorkbook = FsWorkbook()
    metadata_sheet: FsWorksheet = ArcStudy_toMetadataSheet(study, assays)
    doc.AddWorksheet(metadata_sheet)
    if datamap_sheet_1:
        def action(arg: DataMap, study: Any=study, assays: Any=assays, datamap_sheet: Any=datamap_sheet) -> None:
            sheet: FsWorksheet = to_fs_worksheet(arg)
            doc.AddWorksheet(sheet)

        iterate(action, to_array(study.DataMap))

    def action_1(i: int, arg_1: ArcTable, study: Any=study, assays: Any=assays, datamap_sheet: Any=datamap_sheet) -> None:
        sheet_1: FsWorksheet = to_fs_worksheet_1(i, arg_1)
        doc.AddWorksheet(sheet_1)

    iterate_indexed(action_1, study.Tables)
    return doc


def ARCtrl_ArcStudy__ArcStudy_ToFsWorkbook_257FC1F0(this: ArcStudy, assays: FSharpList[ArcAssay] | None=None, datamap_sheet: bool | None=None) -> FsWorkbook:
    return ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(this, assays, datamap_sheet)


__all__ = ["ArcStudy_toMetadataSheet", "ArcStudy_fromRows", "ArcStudy_fromMetadataSheet", "ArcStudy_toMetadataCollection", "ArcStudy_fromMetadataCollection", "ArcStudy_isMetadataSheetName", "ArcStudy_isMetadataSheet", "ArcStudy_tryGetMetadataSheet", "ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D", "ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522", "ARCtrl_ArcStudy__ArcStudy_ToFsWorkbook_257FC1F0"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.int32 import parse
from ...fable_library.list import (length, empty, FSharpList, choose, collect, append, try_find, map as map_1, try_pick, zip)
from ...fable_library.option import (default_arg, bind, value as value_4, map)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, list_type, record_type)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.seq2 import List_distinct
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ...fable_library.util import (equals, safe_hash)
from ..comment import (Comment, Comment_reflection)
from ..data import Data
from ..Helper.collections_ import Option_fromValueWithDefault
from ..Helper.regex import ActivePatterns__007CRegex_007C__007C
from ..ontology_annotation import OntologyAnnotation
from .factor import Factor
from .factor_value import FactorValue
from .material import Material
from .material_attribute import MaterialAttribute
from .material_attribute_value import MaterialAttributeValue
from .process_input import (ProcessInput, ProcessInput_reflection, ProcessInput_tryGetCharacteristicValues_5B3D5BA9, ProcessInput_trySource_5B3D5BA9, ProcessInput_tryData_5B3D5BA9, ProcessInput_trySample_5B3D5BA9, ProcessInput_tryMaterial_5B3D5BA9)
from .process_output import (ProcessOutput, ProcessOutput_reflection, ProcessOutput_tryGetCharacteristicValues_Z42C11600, ProcessOutput_tryGetFactorValues_Z42C11600, ProcessOutput_tryData_Z42C11600, ProcessOutput_trySample_Z42C11600, ProcessOutput_tryMaterial_Z42C11600)
from .process_parameter_value import (ProcessParameterValue, ProcessParameterValue_reflection)
from .protocol import (Protocol, Protocol_reflection)
from .protocol_parameter import ProtocolParameter
from .sample import Sample
from .source import Source

def _expr774() -> TypeInfo:
    return record_type("ARCtrl.Process.Process", [], Process, lambda: [("ID", option_type(string_type)), ("Name", option_type(string_type)), ("ExecutesProtocol", option_type(Protocol_reflection())), ("ParameterValues", option_type(list_type(ProcessParameterValue_reflection()))), ("Performer", option_type(string_type)), ("Date", option_type(string_type)), ("PreviousProcess", option_type(Process_reflection())), ("NextProcess", option_type(Process_reflection())), ("Inputs", option_type(list_type(ProcessInput_reflection()))), ("Outputs", option_type(list_type(ProcessOutput_reflection()))), ("Comments", option_type(list_type(Comment_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Process(Record):
    ID: str | None
    Name: str | None
    ExecutesProtocol: Protocol | None
    ParameterValues: FSharpList[ProcessParameterValue] | None
    Performer: str | None
    Date: str | None
    PreviousProcess: Process | None
    NextProcess: Process | None
    Inputs: FSharpList[ProcessInput] | None
    Outputs: FSharpList[ProcessOutput] | None
    Comments: FSharpList[Comment] | None
    def Print(self, __unit: None=None) -> str:
        this: Process = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Process = self
        input_count: int = length(default_arg(this.Inputs, empty())) or 0
        output_count: int = length(default_arg(this.Outputs, empty())) or 0
        param_count: int = length(default_arg(this.ParameterValues, empty())) or 0
        name: str = default_arg(this.Name, "Unnamed Process")
        return to_text(printf("%s [%i Inputs -> %i Params -> %i Outputs]"))(name)(input_count)(param_count)(output_count)


Process_reflection = _expr774

def Process_make(id: str | None=None, name: str | None=None, executes_protocol: Protocol | None=None, parameter_values: FSharpList[ProcessParameterValue] | None=None, performer: str | None=None, date: str | None=None, previous_process: Process | None=None, next_process: Process | None=None, inputs: FSharpList[ProcessInput] | None=None, outputs: FSharpList[ProcessOutput] | None=None, comments: FSharpList[Comment] | None=None) -> Process:
    return Process(id, name, executes_protocol, parameter_values, performer, date, previous_process, next_process, inputs, outputs, comments)


def Process_create_Z7C1F7FA9(Id: str | None=None, Name: str | None=None, ExecutesProtocol: Protocol | None=None, ParameterValues: FSharpList[ProcessParameterValue] | None=None, Performer: str | None=None, Date: str | None=None, PreviousProcess: Process | None=None, NextProcess: Process | None=None, Inputs: FSharpList[ProcessInput] | None=None, Outputs: FSharpList[ProcessOutput] | None=None, Comments: FSharpList[Comment] | None=None) -> Process:
    return Process_make(Id, Name, ExecutesProtocol, ParameterValues, Performer, Date, PreviousProcess, NextProcess, Inputs, Outputs, Comments)


def Process_get_empty(__unit: None=None) -> Process:
    return Process_create_Z7C1F7FA9()


def Process_composeName(process_name_root: str, i: int) -> str:
    return ((("" + process_name_root) + "_") + str(i)) + ""


def Process_decomposeName_Z721C83C5(name: str) -> tuple[str, int | None]:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("(?<name>.+)_(?<num>\\d+)", name)
    if active_pattern_result is not None:
        r: Any = active_pattern_result
        return (get_item(groups(r), "name") or "", parse(get_item(groups(r), "num") or "", 511, False, 32))

    else: 
        return (name, None)



def Process_tryGetProtocolName_763471FF(p: Process) -> str | None:
    def binder(p_1: Protocol, p: Any=p) -> str | None:
        return p_1.Name

    return bind(binder, p.ExecutesProtocol)


def Process_getProtocolName_763471FF(p: Process) -> str:
    def binder(p_1: Protocol, p: Any=p) -> str | None:
        return p_1.Name

    return value_4(bind(binder, p.ExecutesProtocol))


def Process_getParameterValues_763471FF(p: Process) -> FSharpList[ProcessParameterValue]:
    return default_arg(p.ParameterValues, empty())


def Process_getParameters_763471FF(p: Process) -> FSharpList[ProtocolParameter]:
    def chooser(pv: ProcessParameterValue, p: Any=p) -> ProtocolParameter | None:
        return pv.Category

    return choose(chooser, Process_getParameterValues_763471FF(p))


def Process_getInputCharacteristicValues_763471FF(p: Process) -> FSharpList[MaterialAttributeValue]:
    match_value: FSharpList[ProcessInput] | None = p.Inputs
    if match_value is None:
        return empty()

    else: 
        def mapping(inp: ProcessInput, p: Any=p) -> FSharpList[MaterialAttributeValue]:
            return default_arg(ProcessInput_tryGetCharacteristicValues_5B3D5BA9(inp), empty())

        class ObjectExpr775:
            @property
            def Equals(self) -> Callable[[MaterialAttributeValue, MaterialAttributeValue], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[MaterialAttributeValue], int]:
                return safe_hash

        return List_distinct(collect(mapping, match_value), ObjectExpr775())



def Process_getOutputCharacteristicValues_763471FF(p: Process) -> FSharpList[MaterialAttributeValue]:
    match_value: FSharpList[ProcessOutput] | None = p.Outputs
    if match_value is None:
        return empty()

    else: 
        def mapping(out: ProcessOutput, p: Any=p) -> FSharpList[MaterialAttributeValue]:
            return default_arg(ProcessOutput_tryGetCharacteristicValues_Z42C11600(out), empty())

        class ObjectExpr776:
            @property
            def Equals(self) -> Callable[[MaterialAttributeValue, MaterialAttributeValue], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[MaterialAttributeValue], int]:
                return safe_hash

        return List_distinct(collect(mapping, match_value), ObjectExpr776())



def Process_getCharacteristicValues_763471FF(p: Process) -> FSharpList[MaterialAttributeValue]:
    class ObjectExpr777:
        @property
        def Equals(self) -> Callable[[MaterialAttributeValue, MaterialAttributeValue], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[MaterialAttributeValue], int]:
            return safe_hash

    return List_distinct(append(Process_getInputCharacteristicValues_763471FF(p), Process_getOutputCharacteristicValues_763471FF(p)), ObjectExpr777())


def Process_getCharacteristics_763471FF(p: Process) -> FSharpList[MaterialAttribute]:
    def chooser(cv: MaterialAttributeValue, p: Any=p) -> MaterialAttribute | None:
        return cv.Category

    class ObjectExpr778:
        @property
        def Equals(self) -> Callable[[MaterialAttribute, MaterialAttribute], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[MaterialAttribute], int]:
            return safe_hash

    return List_distinct(choose(chooser, Process_getCharacteristicValues_763471FF(p)), ObjectExpr778())


def Process_getFactorValues_763471FF(p: Process) -> FSharpList[FactorValue]:
    def mapping(arg: ProcessOutput, p: Any=p) -> FSharpList[FactorValue]:
        return default_arg(ProcessOutput_tryGetFactorValues_Z42C11600(arg), empty())

    class ObjectExpr779:
        @property
        def Equals(self) -> Callable[[FactorValue, FactorValue], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[FactorValue], int]:
            return safe_hash

    return List_distinct(collect(mapping, default_arg(p.Outputs, empty())), ObjectExpr779())


def Process_getFactors_763471FF(p: Process) -> FSharpList[Factor]:
    def chooser(fv: FactorValue, p: Any=p) -> Factor | None:
        return fv.Category

    class ObjectExpr780:
        @property
        def Equals(self) -> Callable[[Factor, Factor], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Factor], int]:
            return safe_hash

    return List_distinct(choose(chooser, Process_getFactorValues_763471FF(p)), ObjectExpr780())


def Process_getUnits_763471FF(p: Process) -> FSharpList[OntologyAnnotation]:
    def chooser(cv: MaterialAttributeValue, p: Any=p) -> OntologyAnnotation | None:
        return cv.Unit

    def chooser_1(pv: ProcessParameterValue, p: Any=p) -> OntologyAnnotation | None:
        return pv.Unit

    def chooser_2(fv: FactorValue, p: Any=p) -> OntologyAnnotation | None:
        return fv.Unit

    return append(choose(chooser, Process_getCharacteristicValues_763471FF(p)), append(choose(chooser_1, Process_getParameterValues_763471FF(p)), choose(chooser_2, Process_getFactorValues_763471FF(p))))


def Process_tryGetInputsWithParameterBy(predicate: Callable[[ProtocolParameter], bool], p: Process) -> FSharpList[tuple[ProcessInput, ProcessParameterValue]] | None:
    match_value: FSharpList[ProcessParameterValue] | None = p.ParameterValues
    if match_value is None:
        return None

    else: 
        def predicate_1(pv: ProcessParameterValue, predicate: Any=predicate, p: Any=p) -> bool:
            return predicate(default_arg(pv.Category, ProtocolParameter.empty()))

        match_value_1: ProcessParameterValue | None = try_find(predicate_1, match_value)
        if match_value_1 is None:
            return None

        else: 
            param_value: ProcessParameterValue = match_value_1
            def mapping_1(list_2: FSharpList[ProcessInput], predicate: Any=predicate, p: Any=p) -> FSharpList[tuple[ProcessInput, ProcessParameterValue]]:
                def mapping(i: ProcessInput, list_2: Any=list_2) -> tuple[ProcessInput, ProcessParameterValue]:
                    return (i, param_value)

                return map_1(mapping, list_2)

            return map(mapping_1, p.Inputs)




def Process_tryGetOutputsWithParameterBy(predicate: Callable[[ProtocolParameter], bool], p: Process) -> FSharpList[tuple[ProcessOutput, ProcessParameterValue]] | None:
    match_value: FSharpList[ProcessParameterValue] | None = p.ParameterValues
    if match_value is None:
        return None

    else: 
        def predicate_1(pv: ProcessParameterValue, predicate: Any=predicate, p: Any=p) -> bool:
            return predicate(default_arg(pv.Category, ProtocolParameter.empty()))

        match_value_1: ProcessParameterValue | None = try_find(predicate_1, match_value)
        if match_value_1 is None:
            return None

        else: 
            param_value: ProcessParameterValue = match_value_1
            def mapping_1(list_2: FSharpList[ProcessOutput], predicate: Any=predicate, p: Any=p) -> FSharpList[tuple[ProcessOutput, ProcessParameterValue]]:
                def mapping(i: ProcessOutput, list_2: Any=list_2) -> tuple[ProcessOutput, ProcessParameterValue]:
                    return (i, param_value)

                return map_1(mapping, list_2)

            return map(mapping_1, p.Outputs)




def Process_tryGetInputsWithCharacteristicBy(predicate: Callable[[MaterialAttribute], bool], p: Process) -> FSharpList[tuple[ProcessInput, MaterialAttributeValue]] | None:
    match_value: FSharpList[ProcessInput] | None = p.Inputs
    if match_value is None:
        return None

    else: 
        def chooser_1(i: ProcessInput, predicate: Any=predicate, p: Any=p) -> tuple[ProcessInput, MaterialAttributeValue] | None:
            def chooser(mv: MaterialAttributeValue, i: Any=i) -> tuple[ProcessInput, MaterialAttributeValue] | None:
                match_value_1: MaterialAttribute | None = mv.Category
                (pattern_matching_result,) = (None,)
                if match_value_1 is not None:
                    if predicate(match_value_1):
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1

                if pattern_matching_result == 0:
                    return (i, mv)

                elif pattern_matching_result == 1:
                    return None


            return try_pick(chooser, default_arg(ProcessInput_tryGetCharacteristicValues_5B3D5BA9(i), empty()))

        return Option_fromValueWithDefault(empty(), choose(chooser_1, match_value))



def Process_tryGetOutputsWithCharacteristicBy(predicate: Callable[[MaterialAttribute], bool], p: Process) -> FSharpList[tuple[ProcessOutput, MaterialAttributeValue]] | None:
    match_value: FSharpList[ProcessInput] | None = p.Inputs
    match_value_1: FSharpList[ProcessOutput] | None = p.Outputs
    (pattern_matching_result, is_, os) = (None, None, None)
    if match_value is not None:
        if match_value_1 is not None:
            pattern_matching_result = 0
            is_ = match_value
            os = match_value_1

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        def chooser_1(tupled_arg: tuple[ProcessInput, ProcessOutput], predicate: Any=predicate, p: Any=p) -> tuple[ProcessOutput, MaterialAttributeValue] | None:
            def chooser(mv: MaterialAttributeValue, tupled_arg: Any=tupled_arg) -> tuple[ProcessOutput, MaterialAttributeValue] | None:
                match_value_1: MaterialAttribute | None = mv.Category
                (pattern_matching_result_1,) = (None,)
                if match_value_1 is not None:
                    if predicate(match_value_1):
                        pattern_matching_result_1 = 0

                    else: 
                        pattern_matching_result_1 = 1


                else: 
                    pattern_matching_result_1 = 1

                if pattern_matching_result_1 == 0:
                    return (tupled_arg[1], mv)

                elif pattern_matching_result_1 == 1:
                    return None


            return try_pick(chooser, default_arg(ProcessInput_tryGetCharacteristicValues_5B3D5BA9(tupled_arg[0]), empty()))

        return Option_fromValueWithDefault(empty(), choose(chooser_1, zip(is_, os)))

    elif pattern_matching_result == 1:
        return None



def Process_tryGetOutputsWithFactorBy(predicate: Callable[[Factor], bool], p: Process) -> FSharpList[tuple[ProcessOutput, FactorValue]] | None:
    match_value: FSharpList[ProcessOutput] | None = p.Outputs
    if match_value is None:
        return None

    else: 
        def chooser_1(o: ProcessOutput, predicate: Any=predicate, p: Any=p) -> tuple[ProcessOutput, FactorValue] | None:
            def chooser(mv: FactorValue, o: Any=o) -> tuple[ProcessOutput, FactorValue] | None:
                match_value_1: Factor | None = mv.Category
                (pattern_matching_result,) = (None,)
                if match_value_1 is not None:
                    if predicate(match_value_1):
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1

                if pattern_matching_result == 0:
                    return (o, mv)

                elif pattern_matching_result == 1:
                    return None


            return try_pick(chooser, default_arg(ProcessOutput_tryGetFactorValues_Z42C11600(o), empty()))

        return Option_fromValueWithDefault(empty(), choose(chooser_1, match_value))



def Process_getSources_763471FF(p: Process) -> FSharpList[Source]:
    def chooser(pi: ProcessInput, p: Any=p) -> Source | None:
        return ProcessInput_trySource_5B3D5BA9(pi)

    class ObjectExpr781:
        @property
        def Equals(self) -> Callable[[Source, Source], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Source], int]:
            return safe_hash

    return List_distinct(choose(chooser, default_arg(p.Inputs, empty())), ObjectExpr781())


def Process_getData_763471FF(p: Process) -> FSharpList[Data]:
    def chooser(pi: ProcessInput, p: Any=p) -> Data | None:
        return ProcessInput_tryData_5B3D5BA9(pi)

    def chooser_1(po: ProcessOutput, p: Any=p) -> Data | None:
        return ProcessOutput_tryData_Z42C11600(po)

    class ObjectExpr782:
        @property
        def Equals(self) -> Callable[[Data, Data], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Data], int]:
            return safe_hash

    return List_distinct(append(choose(chooser, default_arg(p.Inputs, empty())), choose(chooser_1, default_arg(p.Outputs, empty()))), ObjectExpr782())


def Process_getSamples_763471FF(p: Process) -> FSharpList[Sample]:
    def chooser(pi: ProcessInput, p: Any=p) -> Sample | None:
        return ProcessInput_trySample_5B3D5BA9(pi)

    def chooser_1(po: ProcessOutput, p: Any=p) -> Sample | None:
        return ProcessOutput_trySample_Z42C11600(po)

    class ObjectExpr783:
        @property
        def Equals(self) -> Callable[[Sample, Sample], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Sample], int]:
            return safe_hash

    return List_distinct(append(choose(chooser, default_arg(p.Inputs, empty())), choose(chooser_1, default_arg(p.Outputs, empty()))), ObjectExpr783())


def Process_getMaterials_763471FF(p: Process) -> FSharpList[Material]:
    def chooser(pi: ProcessInput, p: Any=p) -> Material | None:
        return ProcessInput_tryMaterial_5B3D5BA9(pi)

    def chooser_1(po: ProcessOutput, p: Any=p) -> Material | None:
        return ProcessOutput_tryMaterial_Z42C11600(po)

    class ObjectExpr784:
        @property
        def Equals(self) -> Callable[[Material, Material], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Material], int]:
            return safe_hash

    return List_distinct(append(choose(chooser, default_arg(p.Inputs, empty())), choose(chooser_1, default_arg(p.Outputs, empty()))), ObjectExpr784())


__all__ = ["Process_reflection", "Process_make", "Process_create_Z7C1F7FA9", "Process_get_empty", "Process_composeName", "Process_decomposeName_Z721C83C5", "Process_tryGetProtocolName_763471FF", "Process_getProtocolName_763471FF", "Process_getParameterValues_763471FF", "Process_getParameters_763471FF", "Process_getInputCharacteristicValues_763471FF", "Process_getOutputCharacteristicValues_763471FF", "Process_getCharacteristicValues_763471FF", "Process_getCharacteristics_763471FF", "Process_getFactorValues_763471FF", "Process_getFactors_763471FF", "Process_getUnits_763471FF", "Process_tryGetInputsWithParameterBy", "Process_tryGetOutputsWithParameterBy", "Process_tryGetInputsWithCharacteristicBy", "Process_tryGetOutputsWithCharacteristicBy", "Process_tryGetOutputsWithFactorBy", "Process_getSources_763471FF", "Process_getData_763471FF", "Process_getSamples_763471FF", "Process_getMaterials_763471FF"]




from __future__ import annotations
from ..fable_library.string_ import (to_text, printf, to_console)
from ..fable_openpyxl.openpyxl import CellType
from ..fs_spreadsheet.Cells.fs_cell import DataType

def from_data_tyoe(t: DataType) -> CellType:
    if t.tag == 1:
        return CellType(3)

    elif t.tag == 2:
        return CellType(0)

    elif t.tag == 3:
        return CellType(4)

    elif t.tag == 0:
        return CellType(2)

    else: 
        msg: str = to_text(printf("ValueType \'%A\' is not fully implemented in FsSpreadsheet and is handled as string input."))(t)
        to_console(printf("%s"))(msg)
        return CellType(2)



def to_data_type(t: CellType) -> DataType:
    if t.tag == 3:
        return DataType(1)

    elif t.tag == 0:
        return DataType(2)

    elif t.tag == 1:
        return DataType(2)

    elif t.tag == 4:
        return DataType(3)

    elif t.tag == 2:
        return DataType(0)

    else: 
        msg: str = to_text(printf("ValueType \'%A\' is not fully implemented in FsSpreadsheet and is handled as string input."))(t)
        to_console(printf("%s"))(msg)
        return DataType(0)



__all__ = ["from_data_tyoe", "to_data_type"]




from __future__ import annotations
from typing import Any
from ..fable_library.array_ import (try_find, exists, filter)
from ..fable_library.list import (append, singleton, FSharpList, map, choose, of_array)
from ..fable_library.option import (default_arg, map as map_1)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import length
from ..fable_library.string_ import (join, to_text, printf)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_Z721C83C5)
from ..fable_library.types import (Array, to_string)
from ..fable_library.util import (equals, ignore)
from .comment import Comment
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq, hash_1)
from .ontology_annotation import OntologyAnnotation

def _expr741() -> TypeInfo:
    return class_type("ARCtrl.Person", None, Person)


class Person:
    def __init__(self, orcid: str | None=None, last_name: str | None=None, first_name: str | None=None, mid_initials: str | None=None, email: str | None=None, phone: str | None=None, fax: str | None=None, address: str | None=None, affiliation: str | None=None, roles: Array[OntologyAnnotation] | None=None, comments: Array[Comment] | None=None) -> None:
        self._orcid: str | None = orcid
        self._lastName: str | None = last_name
        self._firstName: str | None = first_name
        self._midInitials: str | None = mid_initials
        self._email: str | None = email
        self._phone: str | None = phone
        self._fax: str | None = fax
        self._address: str | None = address
        self._affiliation: str | None = affiliation
        self._roles: Array[OntologyAnnotation] = default_arg(roles, [])
        self._comments: Array[Comment] = default_arg(comments, [])

    @property
    def ORCID(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._orcid

    @ORCID.setter
    def ORCID(self, orcid: str | None=None) -> None:
        this: Person = self
        this._orcid = orcid

    @property
    def FirstName(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._firstName

    @FirstName.setter
    def FirstName(self, first_name: str | None=None) -> None:
        this: Person = self
        this._firstName = first_name

    @property
    def LastName(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._lastName

    @LastName.setter
    def LastName(self, last_name: str | None=None) -> None:
        this: Person = self
        this._lastName = last_name

    @property
    def MidInitials(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._midInitials

    @MidInitials.setter
    def MidInitials(self, mid_initials: str | None=None) -> None:
        this: Person = self
        this._midInitials = mid_initials

    @property
    def Address(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._address

    @Address.setter
    def Address(self, address: str | None=None) -> None:
        this: Person = self
        this._address = address

    @property
    def Affiliation(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._affiliation

    @Affiliation.setter
    def Affiliation(self, affiliation: str | None=None) -> None:
        this: Person = self
        this._affiliation = affiliation

    @property
    def EMail(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._email

    @EMail.setter
    def EMail(self, email: str | None=None) -> None:
        this: Person = self
        this._email = email

    @property
    def Phone(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._phone

    @Phone.setter
    def Phone(self, phone: str | None=None) -> None:
        this: Person = self
        this._phone = phone

    @property
    def Fax(self, __unit: None=None) -> str | None:
        this: Person = self
        return this._fax

    @Fax.setter
    def Fax(self, fax: str | None=None) -> None:
        this: Person = self
        this._fax = fax

    @property
    def Roles(self, __unit: None=None) -> Array[OntologyAnnotation]:
        this: Person = self
        return this._roles

    @Roles.setter
    def Roles(self, roles: Array[OntologyAnnotation]) -> None:
        this: Person = self
        this._roles = roles

    @property
    def Comments(self, __unit: None=None) -> Array[Comment]:
        this: Person = self
        return this._comments

    @Comments.setter
    def Comments(self, comments: Array[Comment]) -> None:
        this: Person = self
        this._comments = comments

    @staticmethod
    def make(orcid: str | None, last_name: str | None, first_name: str | None, mid_initials: str | None, email: str | None, phone: str | None, fax: str | None, address: str | None, affiliation: str | None, roles: Array[OntologyAnnotation], comments: Array[Comment]) -> Person:
        return Person(orcid, last_name, first_name, mid_initials, email, phone, fax, address, affiliation, roles, comments)

    @staticmethod
    def create(orcid: str | None=None, last_name: str | None=None, first_name: str | None=None, mid_initials: str | None=None, email: str | None=None, phone: str | None=None, fax: str | None=None, address: str | None=None, affiliation: str | None=None, roles: Array[OntologyAnnotation] | None=None, comments: Array[Comment] | None=None) -> Person:
        roles_1: Array[OntologyAnnotation] = default_arg(roles, [])
        comments_1: Array[Comment] = default_arg(comments, [])
        return Person.make(orcid, last_name, first_name, mid_initials, email, phone, fax, address, affiliation, roles_1, comments_1)

    @staticmethod
    def empty() -> Person:
        return Person.create()

    @staticmethod
    def try_get_by_full_name(first_name: str, mid_initials: str, last_name: str, persons: Array[Person]) -> Person | None:
        def predicate(p: Person) -> bool:
            if mid_initials == "":
                if equals(p.FirstName, first_name):
                    return equals(p.LastName, last_name)

                else: 
                    return False


            elif equals(p.MidInitials, mid_initials) if equals(p.FirstName, first_name) else False:
                return equals(p.LastName, last_name)

            else: 
                return False


        return try_find(predicate, persons)

    @staticmethod
    def exists_by_full_name(first_name: str, mid_initials: str, last_name: str, persons: Array[Person]) -> bool:
        def _arrow736(p: Person) -> bool:
            return (equals(p.LastName, last_name) if equals(p.FirstName, first_name) else False) if (mid_initials == "") else (equals(p.LastName, last_name) if (equals(p.MidInitials, mid_initials) if equals(p.FirstName, first_name) else False) else False)

        return exists(_arrow736, persons)

    @staticmethod
    def add(persons: FSharpList[Person], person: Person) -> FSharpList[Person]:
        return append(persons, singleton(person))

    @staticmethod
    def remove_by_full_name(first_name: str, mid_initials: str, last_name: str, persons: Array[Person]) -> Array[Person]:
        def _arrow737(p: Person) -> bool:
            return (not (equals(p.LastName, last_name) if equals(p.FirstName, first_name) else False)) if (mid_initials == "") else (not (equals(p.LastName, last_name) if (equals(p.MidInitials, mid_initials) if equals(p.FirstName, first_name) else False) else False))

        return filter(_arrow737, persons)

    def Copy(self, __unit: None=None) -> Person:
        this: Person = self
        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        def f_1(c_1: OntologyAnnotation) -> OntologyAnnotation:
            return c_1.Copy()

        next_roles: Array[OntologyAnnotation] = ResizeArray_map(f_1, this.Roles)
        orcid: str | None = this.ORCID
        last_name: str | None = this.LastName
        first_name: str | None = this.FirstName
        mid_initials: str | None = this.MidInitials
        email: str | None = this.EMail
        phone: str | None = this.Phone
        fax: str | None = this.Fax
        address: str | None = this.Address
        affiliation: str | None = this.Affiliation
        return Person.make(orcid, last_name, first_name, mid_initials, email, phone, fax, address, affiliation, next_roles, next_comments)

    def __hash__(self, __unit: None=None) -> Any:
        this: Person = self
        return box_hash_array([box_hash_option(this.ORCID), box_hash_option(this.LastName), box_hash_option(this.FirstName), box_hash_option(this.MidInitials), box_hash_option(this.EMail), box_hash_option(this.Phone), box_hash_option(this.Fax), box_hash_option(this.Address), box_hash_option(this.Affiliation), box_hash_seq(this.Roles), box_hash_seq(this.Comments)])

    def __eq__(self, obj: Any=None) -> bool:
        this: Person = self
        return hash_1(this) == hash_1(obj)

    def __str__(self, __unit: None=None) -> str:
        this: Person = self
        sb: Any = StringBuilder__ctor()
        ignore(StringBuilder__Append_Z721C83C5(sb, "Person {\n\t"))
        def mapping_1(tupled_arg_1: tuple[str, str]) -> str:
            return to_text(printf("%s = %A"))(tupled_arg_1[0])(tupled_arg_1[1])

        def chooser(tupled_arg: tuple[str, str | None]) -> tuple[str, str] | None:
            def mapping(o: str, tupled_arg: Any=tupled_arg) -> tuple[str, str]:
                return (tupled_arg[0], o)

            return map_1(mapping, tupled_arg[1])

        def _arrow738(__unit: None=None) -> str:
            arg: Array[OntologyAnnotation] = this.Roles
            return to_text(printf("%A"))(arg)

        def _arrow739(__unit: None=None) -> str:
            arg_1: Array[Comment] = this.Comments
            return to_text(printf("%A"))(arg_1)

        ignore(StringBuilder__Append_Z721C83C5(sb, join(",\n\t", map(mapping_1, choose(chooser, of_array([("FirstName", this.FirstName), ("LastName", this.LastName), ("MidInitials", this.MidInitials), ("EMail", this.EMail), ("Phone", this.Phone), ("Address", this.Address), ("Affiliation", this.Affiliation), ("Fax", this.Fax), ("ORCID", this.ORCID), ("Roles", _arrow738() if (length(this.Roles) > 0) else None), ("Comments", _arrow739() if (length(this.Comments) > 0) else None)]))))))
        ignore(StringBuilder__Append_Z721C83C5(sb, "\n}"))
        return to_string(sb)


Person_reflection = _expr741

def Person__ctor_Z2F6491B5(orcid: str | None=None, last_name: str | None=None, first_name: str | None=None, mid_initials: str | None=None, email: str | None=None, phone: str | None=None, fax: str | None=None, address: str | None=None, affiliation: str | None=None, roles: Array[OntologyAnnotation] | None=None, comments: Array[Comment] | None=None) -> Person:
    return Person(orcid, last_name, first_name, mid_initials, email, phone, fax, address, affiliation, roles, comments)


__all__ = ["Person_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_rocrate.ldobject import LDNode
from ...fable_library.int32 import try_parse
from ...fable_library.option import value
from ...fable_library.types import FSharpRef
from ...fable_library.util import int32_to_string

def try_int(str_1: str) -> int | None:
    match_value: tuple[bool, int]
    out_arg: int = 0
    def _arrow3871(__unit: None=None, str_1: Any=str_1) -> int:
        return out_arg

    def _arrow3872(v: int, str_1: Any=str_1) -> None:
        nonlocal out_arg
        out_arg = v or 0

    match_value = (try_parse(str_1, 511, False, 32, FSharpRef(_arrow3871, _arrow3872)), out_arg)
    if match_value[0]:
        return match_value[1]

    else: 
        return None



order_name: str = "columnIndex"

def try_get_index(node: LDNode) -> int | None:
    match_value: Any | None = node.TryGetPropertyAsSingleton(order_name)
    (pattern_matching_result, ci) = (None, None)
    if match_value is not None:
        if str(type(value(match_value))) == "<class \'str\'>":
            pattern_matching_result = 0
            ci = value(match_value)

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return try_int(ci)

    elif pattern_matching_result == 1:
        return None



def set_index(node: LDNode, index: int) -> None:
    node.SetProperty(order_name, int32_to_string(index))


def ARCtrl_ROCrate_LDNode__LDNode_GetColumnIndex(this: LDNode) -> int:
    return value(try_get_index(this))


def ARCtrl_ROCrate_LDNode__LDNode_TryGetColumnIndex(this: LDNode) -> int | None:
    return try_get_index(this)


def ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4(this: LDNode, index: int) -> None:
    set_index(this, index)


__all__ = ["try_int", "order_name", "try_get_index", "set_index", "ARCtrl_ROCrate_LDNode__LDNode_GetColumnIndex", "ARCtrl_ROCrate_LDNode__LDNode_TryGetColumnIndex", "ARCtrl_ROCrate_LDNode__LDNode_SetColumnIndex_Z524259A4"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.list import (map, choose, of_array)
from ..fable_library.option import (default_arg, map as map_1)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import length
from ..fable_library.string_ import (join, to_text, printf)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_Z721C83C5)
from ..fable_library.types import (Array, to_string)
from ..fable_library.util import (safe_hash, ignore)
from .comment import Comment
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq)
from .ontology_annotation import OntologyAnnotation

def _expr735() -> TypeInfo:
    return class_type("ARCtrl.Publication", None, Publication)


class Publication:
    def __init__(self, pub_med_id: str | None=None, doi: str | None=None, authors: str | None=None, title: str | None=None, status: OntologyAnnotation | None=None, comments: Array[Comment] | None=None) -> None:
        self._pubMedID: str | None = pub_med_id
        self._doi: str | None = doi
        self._authors: str | None = authors
        self._title: str | None = title
        self._status: OntologyAnnotation | None = status
        self._comments: Array[Comment] = default_arg(comments, [])

    @property
    def PubMedID(self, __unit: None=None) -> str | None:
        this: Publication = self
        return this._pubMedID

    @PubMedID.setter
    def PubMedID(self, pub_med_id: str | None=None) -> None:
        this: Publication = self
        this._pubMedID = pub_med_id

    @property
    def DOI(self, __unit: None=None) -> str | None:
        this: Publication = self
        return this._doi

    @DOI.setter
    def DOI(self, doi: str | None=None) -> None:
        this: Publication = self
        this._doi = doi

    @property
    def Authors(self, __unit: None=None) -> str | None:
        this: Publication = self
        return this._authors

    @Authors.setter
    def Authors(self, authors: str | None=None) -> None:
        this: Publication = self
        this._authors = authors

    @property
    def Title(self, __unit: None=None) -> str | None:
        this: Publication = self
        return this._title

    @Title.setter
    def Title(self, title: str | None=None) -> None:
        this: Publication = self
        this._title = title

    @property
    def Status(self, __unit: None=None) -> OntologyAnnotation | None:
        this: Publication = self
        return this._status

    @Status.setter
    def Status(self, status: OntologyAnnotation | None=None) -> None:
        this: Publication = self
        this._status = status

    @property
    def Comments(self, __unit: None=None) -> Array[Comment]:
        this: Publication = self
        return this._comments

    @Comments.setter
    def Comments(self, comments: Array[Comment]) -> None:
        this: Publication = self
        this._comments = comments

    @staticmethod
    def make(pub_med_id: str | None, doi: str | None, authors: str | None, title: str | None, status: OntologyAnnotation | None, comments: Array[Comment]) -> Publication:
        return Publication(pub_med_id, doi, authors, title, status, comments)

    @staticmethod
    def create(pub_med_id: str | None=None, doi: str | None=None, authors: str | None=None, title: str | None=None, status: OntologyAnnotation | None=None, comments: Array[Comment] | None=None) -> Publication:
        comments_1: Array[Comment] = default_arg(comments, [])
        return Publication.make(pub_med_id, doi, authors, title, status, comments_1)

    @staticmethod
    def empty(__unit: None=None) -> Publication:
        return Publication.create()

    def Copy(self, __unit: None=None) -> Publication:
        this: Publication = self
        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        pub_med_id: str | None = this.PubMedID
        doi: str | None = this.DOI
        authors: str | None = this.Authors
        title: str | None = this.Title
        status: OntologyAnnotation | None = this.Status
        return Publication.make(pub_med_id, doi, authors, title, status, next_comments)

    def __hash__(self, __unit: None=None) -> Any:
        this: Publication = self
        return box_hash_array([box_hash_option(this.DOI), box_hash_option(this.Title), box_hash_option(this.Authors), box_hash_option(this.PubMedID), box_hash_option(this.Status), box_hash_seq(this.Comments)])

    def __eq__(self, obj: Any=None) -> bool:
        this: Publication = self
        return (safe_hash(this) == safe_hash(obj)) if isinstance(obj, Publication) else False

    def __str__(self, __unit: None=None) -> str:
        this: Publication = self
        sb: Any = StringBuilder__ctor()
        ignore(StringBuilder__Append_Z721C83C5(sb, "Publication {\n\t"))
        def mapping_2(tupled_arg_1: tuple[str, str]) -> str:
            return to_text(printf("%s = %A"))(tupled_arg_1[0])(tupled_arg_1[1])

        def chooser(tupled_arg: tuple[str, str | None]) -> tuple[str, str] | None:
            def mapping_1(o: str, tupled_arg: Any=tupled_arg) -> tuple[str, str]:
                return (tupled_arg[0], o)

            return map_1(mapping_1, tupled_arg[1])

        def _arrow733(__unit: None=None) -> str | None:
            option: OntologyAnnotation | None = this.Status
            def _arrow732(__unit: None=None) -> Callable[[OntologyAnnotation], str]:
                clo: Callable[[OntologyAnnotation], str] = to_text(printf("%A"))
                return clo

            return map_1(_arrow732(), option)

        def _arrow734(__unit: None=None) -> str:
            arg_1: Array[Comment] = this.Comments
            return to_text(printf("%A"))(arg_1)

        ignore(StringBuilder__Append_Z721C83C5(sb, join(",\n\t", map(mapping_2, choose(chooser, of_array([("PubMedID", this.PubMedID), ("DOI", this.DOI), ("Authors", this.Authors), ("Title", this.Title), ("Status", _arrow733()), ("Comments", _arrow734() if (length(this.Comments) > 0) else None)]))))))
        ignore(StringBuilder__Append_Z721C83C5(sb, "\n}"))
        return to_string(sb)


Publication_reflection = _expr735

def Publication__ctor_Z645CED36(pub_med_id: str | None=None, doi: str | None=None, authors: str | None=None, title: str | None=None, status: OntologyAnnotation | None=None, comments: Array[Comment] | None=None) -> Publication:
    return Publication(pub_med_id, doi, authors, title, status, comments)


__all__ = ["Publication_reflection"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ..arctrl_core.arc_types import ArcWorkflow
from ..arctrl_core.comment import Comment
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_core.Process.component import Component
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IRequiredGetter, string, IOptionalGetter, resize_array, IGetters)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_4, decoder as decoder_4)
from .DataMap.data_map import (encoder as encoder_1, decoder as decoder_2, encoder_compressed as encoder_compressed_1, decoder_compressed as decoder_compressed_1)
from .encode import (try_include, try_include_seq)
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)
from .person import (encoder as encoder_3, decoder as decoder_3)
from .Process.component import (encoder as encoder_2, decoder as decoder_1)

__A_ = TypeVar("__A_")

def encoder(workflow: ArcWorkflow) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], workflow: Any=workflow) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2994(__unit: None=None, workflow: Any=workflow) -> IEncodable:
        value: str = workflow.Identifier
        class ObjectExpr2993(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2993()

    def _arrow2995(oa: OntologyAnnotation, workflow: Any=workflow) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2997(value_1: str, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr2996(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr2996()

    def _arrow2999(value_3: str, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr2998(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr2998()

    def _arrow3001(value_5: str, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3000(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3000()

    def _arrow3003(value_7: str, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3002(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3002()

    def _arrow3004(dm: DataMap, workflow: Any=workflow) -> IEncodable:
        return encoder_1(dm)

    def _arrow3006(value_9: str, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3005(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3005()

    def _arrow3007(oa_1: OntologyAnnotation, workflow: Any=workflow) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3008(value_11: Component, workflow: Any=workflow) -> IEncodable:
        return encoder_2(value_11)

    def _arrow3009(person: Person, workflow: Any=workflow) -> IEncodable:
        return encoder_3(person)

    def _arrow3010(comment: Comment, workflow: Any=workflow) -> IEncodable:
        return encoder_4(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow2994()), try_include("WorkflowType", _arrow2995, workflow.WorkflowType), try_include("Title", _arrow2997, workflow.Title), try_include("URI", _arrow2999, workflow.URI), try_include("Description", _arrow3001, workflow.Description), try_include("Version", _arrow3003, workflow.Version), try_include("DataMap", _arrow3004, workflow.DataMap), try_include_seq("SubWorkflowIdentifiers", _arrow3006, workflow.SubWorkflowIdentifiers), try_include_seq("Parameters", _arrow3007, workflow.Parameters), try_include_seq("Components", _arrow3008, workflow.Components), try_include_seq("Contacts", _arrow3009, workflow.Contacts), try_include_seq("Comments", _arrow3010, workflow.Comments)]))
    class ObjectExpr3011(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], workflow: Any=workflow) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3011()


def _arrow3024(get: IGetters) -> ArcWorkflow:
    def _arrow3012(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("Identifier", string)

    def _arrow3013(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("Title", string)

    def _arrow3014(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("Description", string)

    def _arrow3015(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("WorkflowType", OntologyAnnotation_decoder)

    def _arrow3016(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("URI", string)

    def _arrow3017(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("Version", string)

    def _arrow3018(__unit: None=None) -> Array[str] | None:
        arg_13: Decoder_1[Array[str]] = resize_array(string)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("SubWorkflowIdentifiers", arg_13)

    def _arrow3019(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("Parameters", arg_15)

    def _arrow3020(__unit: None=None) -> Array[Component] | None:
        arg_17: Decoder_1[Array[Component]] = resize_array(decoder_1)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("Components", arg_17)

    def _arrow3021(__unit: None=None) -> DataMap | None:
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("DataMap", decoder_2)

    def _arrow3022(__unit: None=None) -> Array[Person] | None:
        arg_21: Decoder_1[Array[Person]] = resize_array(decoder_3)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("Contacts", arg_21)

    def _arrow3023(__unit: None=None) -> Array[Comment] | None:
        arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_4)
        object_arg_11: IOptionalGetter = get.Optional
        return object_arg_11.Field("Comments", arg_23)

    return ArcWorkflow.create(_arrow3012(), _arrow3013(), _arrow3014(), _arrow3015(), _arrow3016(), _arrow3017(), _arrow3018(), _arrow3019(), _arrow3020(), _arrow3021(), _arrow3022(), None, _arrow3023())


decoder: Decoder_1[ArcWorkflow] = object(_arrow3024)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, workflow: ArcWorkflow) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3028(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        value: str = workflow.Identifier
        class ObjectExpr3027(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3027()

    def _arrow3029(oa: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow3031(value_1: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3030(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3030()

    def _arrow3033(value_3: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3032(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3032()

    def _arrow3035(value_5: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3034(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3034()

    def _arrow3037(value_7: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3036(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3036()

    def _arrow3038(dm: DataMap, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return encoder_compressed_1(string_table, oa_table, cell_table, dm)

    def _arrow3040(value_9: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        class ObjectExpr3039(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3039()

    def _arrow3041(oa_1: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow3042(value_11: Component, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return encoder_2(value_11)

    def _arrow3043(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return encoder_3(person)

    def _arrow3044(comment: Comment, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> IEncodable:
        return encoder_4(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3028()), try_include("WorkflowType", _arrow3029, workflow.WorkflowType), try_include("Title", _arrow3031, workflow.Title), try_include("URI", _arrow3033, workflow.URI), try_include("Description", _arrow3035, workflow.Description), try_include("Version", _arrow3037, workflow.Version), try_include("DataMap", _arrow3038, workflow.DataMap), try_include_seq("SubWorkflowIdentifiers", _arrow3040, workflow.SubWorkflowIdentifiers), try_include_seq("Parameters", _arrow3041, workflow.Parameters), try_include_seq("Components", _arrow3042, workflow.Components), try_include_seq("Contacts", _arrow3043, workflow.Contacts), try_include_seq("Comments", _arrow3044, workflow.Comments)]))
    class ObjectExpr3045(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, workflow: Any=workflow) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3045()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcWorkflow]:
    def _arrow3058(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcWorkflow:
        def _arrow3046(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("Identifier", string)

        def _arrow3047(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("Title", string)

        def _arrow3048(__unit: None=None) -> str | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("Description", string)

        def _arrow3049(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("WorkflowType", OntologyAnnotation_decoder)

        def _arrow3050(__unit: None=None) -> str | None:
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("URI", string)

        def _arrow3051(__unit: None=None) -> str | None:
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("Version", string)

        def _arrow3052(__unit: None=None) -> Array[str] | None:
            arg_13: Decoder_1[Array[str]] = resize_array(string)
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("SubWorkflowIdentifiers", arg_13)

        def _arrow3053(__unit: None=None) -> Array[OntologyAnnotation] | None:
            arg_15: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("Parameters", arg_15)

        def _arrow3054(__unit: None=None) -> Array[Component] | None:
            arg_17: Decoder_1[Array[Component]] = resize_array(decoder_1)
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("Components", arg_17)

        def _arrow3055(__unit: None=None) -> DataMap | None:
            arg_19: Decoder_1[DataMap] = decoder_compressed_1(string_table, oa_table, cell_table)
            object_arg_9: IOptionalGetter = get.Optional
            return object_arg_9.Field("DataMap", arg_19)

        def _arrow3056(__unit: None=None) -> Array[Person] | None:
            arg_21: Decoder_1[Array[Person]] = resize_array(decoder_3)
            object_arg_10: IOptionalGetter = get.Optional
            return object_arg_10.Field("Contacts", arg_21)

        def _arrow3057(__unit: None=None) -> Array[Comment] | None:
            arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_4)
            object_arg_11: IOptionalGetter = get.Optional
            return object_arg_11.Field("Comments", arg_23)

        return ArcWorkflow.create(_arrow3046(), _arrow3047(), _arrow3048(), _arrow3049(), _arrow3050(), _arrow3051(), _arrow3052(), _arrow3053(), _arrow3054(), _arrow3055(), _arrow3056(), None, _arrow3057())

    return object(_arrow3058)


__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from .fable_modules.arctrl_python.JsonIO.ldobject import ARCtrl_ROCrate_LDGraph__LDGraph_ToROCrateJsonString_71136F3F
from .fable_modules.arctrl_rocrate.ldcontext import LDContext
from .fable_modules.arctrl_rocrate.ldobject import (LDNode, LDNode_reflection, LDRef, LDGraph)
from .fable_modules.arctrl_rocrate.LDTypes.creative_work import LDCreativeWork
from .fable_modules.arctrl_rocrate.LDTypes.dataset import LDDataset
from .fable_modules.arctrl_rocrate.LDTypes.defined_term import LDDefinedTerm
from .fable_modules.arctrl_rocrate.LDTypes.file import LDFile
from .fable_modules.arctrl_rocrate.LDTypes.lab_protocol import LDLabProtocol
from .fable_modules.arctrl_rocrate.LDTypes.organization import LDOrganization
from .fable_modules.arctrl_rocrate.LDTypes.person import LDPerson
from .fable_modules.arctrl_rocrate.rocrate_context import init_v1_2
from .fable_modules.fable_library.date import utc_now
from .fable_modules.fable_library.list import (FSharpList, length)
from .fable_modules.fable_library.option import (value as value_2, default_arg)
from .fable_modules.fable_library.reflection import (TypeInfo, class_type)
from .fable_modules.fable_library.seq import (to_list, delay, collect, append, singleton, map)
from .fable_modules.fable_library.seq2 import List_distinct
from .fable_modules.fable_library.types import (FSharpRef, Array)
from .fable_modules.fable_library.util import (IEnumerable_1, equals, safe_hash)

def _expr4100() -> TypeInfo:
    return class_type("ROCratePCC.CustomResourceDescriptor", None, CustomResourceDescriptor)


class CustomResourceDescriptor:
    def __init__(self, id: str, role: str) -> None:
        self.id: str = id
        self.role: str = role

    @property
    def Id(self, __unit: None=None) -> str:
        this: CustomResourceDescriptor = self
        return this.id

    @property
    def Role(self, __unit: None=None) -> str:
        this: CustomResourceDescriptor = self
        return this.role


CustomResourceDescriptor_reflection = _expr4100

def CustomResourceDescriptor__ctor_Z384F8060(id: str, role: str) -> CustomResourceDescriptor:
    return CustomResourceDescriptor(id, role)


def ResourceDescriptorType__get_ID(this: ResourceDescriptorType) -> str:
    if this == "constraint":
        return "#hasConstraint"

    elif this == "guidance":
        return "#hasGuidance"

    elif this == "example":
        return "#hasExample"

    elif isinstance(this, CustomResourceDescriptor):
        return this.Id

    else: 
        return "#hasSpecification"



def ResourceDescriptorType__get_Role(this: ResourceDescriptorType) -> str:
    if this == "constraint":
        return "http://www.w3.org/ns/dx/prof/role/constraints"

    elif this == "guidance":
        return "http://www.w3.org/ns/dx/prof/role/guidance"

    elif this == "example":
        return "http://www.w3.org/ns/dx/prof/role/example"

    elif isinstance(this, CustomResourceDescriptor):
        return this.Role

    else: 
        return "http://www.w3.org/ns/dx/prof/role/specification"



def _expr4101() -> TypeInfo:
    return class_type("ROCratePCC.Organization", None, Organization, LDNode_reflection())


class Organization(LDNode):
    def __init__(self, name: str, url: str | None=None, address: str | None=None, department: Organization | None=None, orcid: Any | None=None) -> None:
        super().__init__(value_2(url) if (url is not None) else (("#" + name) + ""), [LDOrganization.schema_type()])
        n: FSharpRef[Organization] = FSharpRef(None)
        n.contents = self
        self.init_004042: int = 1
        LDDataset.set_name_as_string(n.contents, name)
        if url is not None:
            LDDataset.set_url_as_string(n.contents, value_2(url))

        if address is not None:
            n.contents.SetProperty("http://schema.org/address", value_2(address))

        if department is not None:
            n.contents.SetProperty("http://schema.org/department", value_2(department))



Organization_reflection = _expr4101

def Organization__ctor_Z4D24BC84(name: str, url: str | None=None, address: str | None=None, department: Organization | None=None, orcid: Any | None=None) -> Organization:
    return Organization(name, url, address, department, orcid)


def _expr4102() -> TypeInfo:
    return class_type("ROCratePCC.Author", None, Author, LDNode_reflection())


class Author(LDNode):
    def __init__(self, orcid: str, name: str | None=None, given_name: str | None=None, family_name: str | None=None, email: str | None=None, affiliation: Organization | None=None) -> None:
        super().__init__(("https://orcid.org/" + orcid) + "", [LDPerson.schema_type()])
        n: FSharpRef[Author] = FSharpRef(None)
        n.contents = self
        self.init_004052_002D1: int = 1
        if name is not None:
            LDDataset.set_name_as_string(n.contents, value_2(name))

        if affiliation is not None:
            LDPerson.set_affiliation(n.contents, value_2(affiliation))

        if given_name is not None:
            LDPerson.set_given_name_as_string(n.contents, value_2(given_name))

        if family_name is not None:
            LDPerson.set_family_name_as_string(n.contents, value_2(family_name))

        if email is not None:
            LDPerson.set_email_as_string(n.contents, value_2(email))



Author_reflection = _expr4102

def Author__ctor_Z56C7F91C(orcid: str, name: str | None=None, given_name: str | None=None, family_name: str | None=None, email: str | None=None, affiliation: Organization | None=None) -> Author:
    return Author(orcid, name, given_name, family_name, email, affiliation)


def _expr4103() -> TypeInfo:
    return class_type("ROCratePCC.UsedType", None, UsedType, LDNode_reflection())


class UsedType(LDNode):
    def __init__(self, iri: str, name: str, term_code: Any | None=None) -> None:
        super().__init__(iri, [LDDefinedTerm.schema_type()])
        n: FSharpRef[UsedType] = FSharpRef(None)
        n.contents = self
        self.init_004065_002D2: int = 1
        LDDataset.set_name_as_string(n.contents, name)
        if term_code is not None:
            n.contents.SetProperty("https://schema.org/termCode", value_2(term_code))



UsedType_reflection = _expr4103

def UsedType__ctor_Z6EEF448(iri: str, name: str, term_code: Any | None=None) -> UsedType:
    return UsedType(iri, name, term_code)


def _expr4104() -> TypeInfo:
    return class_type("ROCratePCC.License", None, License, LDNode_reflection())


class License(LDNode):
    def __init__(self, iri: str, name: str) -> None:
        super().__init__(iri, [LDCreativeWork.schema_type()])
        n: FSharpRef[License] = FSharpRef(None)
        n.contents = self
        self.init_004074_002D3: int = 1
        LDDataset.set_name_as_string(n.contents, name)


License_reflection = _expr4104

def License__ctor_Z384F8060(iri: str, name: str) -> License:
    return License(iri, name)


def _expr4105() -> TypeInfo:
    return class_type("ROCratePCC.TextualResource", None, TextualResource, LDNode_reflection())


class TextualResource(LDNode):
    def __init__(self, name: str, file_path: str, encoding_format: str, root_data_entity_id: str | None=None) -> None:
        super().__init__(file_path, [LDFile.schema_type()])
        n: FSharpRef[TextualResource] = FSharpRef(None)
        n.contents = self
        self.init_004079_002D4: int = 1
        LDDataset.set_name_as_string(n.contents, name)
        LDFile.set_encoding_format_as_string(n.contents, encoding_format)
        if root_data_entity_id is None:
            pass

        else: 
            id: str = root_data_entity_id
            n.contents.SetProperty(LDFile.about(), LDRef(id))



TextualResource_reflection = _expr4105

def TextualResource__ctor_5978D483(name: str, file_path: str, encoding_format: str, root_data_entity_id: str | None=None) -> TextualResource:
    return TextualResource(name, file_path, encoding_format, root_data_entity_id)


def _expr4106() -> TypeInfo:
    return class_type("ROCratePCC.ResourceDescriptor", None, ResourceDescriptor, LDNode_reflection())


class ResourceDescriptor(LDNode):
    def __init__(self, textual_resources: Array[TextualResource], resource_descriptor_type: ResourceDescriptorType) -> None:
        super().__init__(ResourceDescriptorType__get_ID(resource_descriptor_type), ["http://www.w3.org/ns/dx/prof/ResourceDescriptor"])
        n: FSharpRef[ResourceDescriptor] = FSharpRef(None)
        n.contents = self
        self.init_004090_002D5: int = 1
        n.contents.SetProperty("http://www.w3.org/ns/dx/prof/hasRole", LDRef(ResourceDescriptorType__get_Role(resource_descriptor_type)))
        n.contents.SetProperty("http://www.w3.org/ns/dx/prof/hasArtifact", textual_resources)


ResourceDescriptor_reflection = _expr4106

def ResourceDescriptor__ctor_40E96EF7(textual_resources: Array[TextualResource], resource_descriptor_type: ResourceDescriptorType) -> ResourceDescriptor:
    return ResourceDescriptor(textual_resources, resource_descriptor_type)


def _expr4107() -> TypeInfo:
    return class_type("ROCratePCC.Specification", None, Specification, ResourceDescriptor_reflection())


class Specification(ResourceDescriptor):
    def __init__(self, textual_resources: Array[TextualResource]) -> None:
        super().__init__(textual_resources, "specification")
        self.init_004099_002D6: int = 1


Specification_reflection = _expr4107

def Specification__ctor_Z715ED06F(textual_resources: Array[TextualResource]) -> Specification:
    return Specification(textual_resources)


def _expr4108() -> TypeInfo:
    return class_type("ROCratePCC.Constraint", None, Constraint, ResourceDescriptor_reflection())


class Constraint(ResourceDescriptor):
    def __init__(self, textual_resources: Array[TextualResource]) -> None:
        super().__init__(textual_resources, "constraint")
        self.init_0040103_002D7: int = 1


Constraint_reflection = _expr4108

def Constraint__ctor_Z715ED06F(textual_resources: Array[TextualResource]) -> Constraint:
    return Constraint(textual_resources)


def _expr4109() -> TypeInfo:
    return class_type("ROCratePCC.Guidance", None, Guidance, ResourceDescriptor_reflection())


class Guidance(ResourceDescriptor):
    def __init__(self, textual_resources: Array[TextualResource]) -> None:
        super().__init__(textual_resources, "guidance")
        self.init_0040107_002D8: int = 1


Guidance_reflection = _expr4109

def Guidance__ctor_Z715ED06F(textual_resources: Array[TextualResource]) -> Guidance:
    return Guidance(textual_resources)


def _expr4110() -> TypeInfo:
    return class_type("ROCratePCC.Example", None, Example, ResourceDescriptor_reflection())


class Example(ResourceDescriptor):
    def __init__(self, textual_resources: Array[TextualResource]) -> None:
        super().__init__(textual_resources, "example")
        self.init_0040111_002D9: int = 1


Example_reflection = _expr4110

def Example__ctor_Z715ED06F(textual_resources: Array[TextualResource]) -> Example:
    return Example(textual_resources)


def _expr4118() -> TypeInfo:
    return class_type("ROCratePCC.RootDataEntity", None, RootDataEntity, LDNode_reflection())


class RootDataEntity(LDNode):
    def __init__(self, id: str, name: str, description: str, license: License, authors: Array[Author], version: str | None=None, keywords: Array[str] | None=None, used_types: Array[UsedType] | None=None, resource_descriptors: Array[ResourceDescriptor] | None=None, data_published: Any | None=None, publisher: Organization | None=None) -> None:
        super().__init__(id, [LDDataset.schema_type(), "http://www.w3.org/ns/dx/prof/Profile"])
        n: FSharpRef[RootDataEntity] = FSharpRef(None)
        n.contents = self
        self.init_0040115_002D10: int = 1
        def _arrow4112(__unit: None=None) -> IEnumerable_1[LDNode]:
            def _arrow4111(rd: ResourceDescriptor) -> IEnumerable_1[LDNode]:
                return rd.GetPropertyNodes("http://www.w3.org/ns/dx/prof/hasArtifact")

            return collect(_arrow4111, default_arg(resource_descriptors, []))

        textual_resources: Array[LDNode] = list(to_list(delay(_arrow4112)))
        def _arrow4116(__unit: None=None) -> IEnumerable_1[LDNode]:
            def _arrow4115(tr: LDNode) -> IEnumerable_1[LDNode]:
                def _arrow4114(__unit: None=None) -> IEnumerable_1[LDNode]:
                    def _arrow4113(ut: UsedType) -> LDNode:
                        return ut

                    return map(_arrow4113, default_arg(used_types, []))

                return append(singleton(tr), delay(_arrow4114))

            return collect(_arrow4115, textual_resources)

        class ObjectExpr4117:
            @property
            def Equals(self) -> Callable[[LDNode, LDNode], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[LDNode], int]:
                return safe_hash

        has_parts: FSharpList[LDNode] = List_distinct(to_list(delay(_arrow4116)), ObjectExpr4117())
        LDDataset.set_license_as_creative_work(n.contents, license)
        LDDataset.set_name_as_string(n.contents, name)
        LDDataset.set_description_as_string(n.contents, description)
        n.contents.SetProperty("http://schema.org/author", authors)
        if keywords is not None:
            n.contents.SetProperty("http://schema.org/keywords", value_2(keywords))

        if version is not None:
            LDLabProtocol.set_version_as_string(n.contents, value_2(version))

        if length(has_parts) > 0:
            LDDataset.set_has_parts(n.contents, list(has_parts))

        if resource_descriptors is not None:
            n.contents.SetProperty("http://www.w3.org/ns/dx/prof/hasResource", value_2(resource_descriptors))

        if data_published is not None:
            LDDataset.set_date_published_as_date_time(n.contents, value_2(data_published))

        else: 
            LDDataset.set_date_published_as_date_time(n.contents, utc_now())

        if publisher is not None:
            n.contents.SetProperty("http://schema.org/publisher", value_2(publisher))



RootDataEntity_reflection = _expr4118

def RootDataEntity__ctor_64728AC0(id: str, name: str, description: str, license: License, authors: Array[Author], version: str | None=None, keywords: Array[str] | None=None, used_types: Array[UsedType] | None=None, resource_descriptors: Array[ResourceDescriptor] | None=None, data_published: Any | None=None, publisher: Organization | None=None) -> RootDataEntity:
    return RootDataEntity(id, name, description, license, authors, version, keywords, used_types, resource_descriptors, data_published, publisher)


def _expr4128() -> TypeInfo:
    return class_type("ROCratePCC.Profile", None, Profile, LDNode_reflection())


class Profile(LDNode):
    def __init__(self, root_data_entity: RootDataEntity, license: License | None=None, ro_crate_spec: str | None=None) -> None:
        super().__init__("ro-crate-metadata.json", [LDCreativeWork.schema_type()])
        n: FSharpRef[Profile] = FSharpRef(None)
        n.contents = self
        self.init_0040142_002D11: int = 1
        LDDataset.set_abouts(n.contents, [root_data_entity])
        if license is not None:
            LDDataset.set_license_as_creative_work(n.contents, value_2(license))

        ro_crate_spec_1: str = default_arg(ro_crate_spec, "https://w3id.org/ro/crate/1.2")
        root_data_entity.SetProperty("http://www.w3.org/ns/dx/prof/isProfileOf", ro_crate_spec_1)
        n.contents.SetProperty("http://purl.org/dc/terms/conformsTo", ro_crate_spec_1)

    def ToROCrateJsonString(self, spaces: int | None=None) -> str:
        this: Profile = self
        context: LDContext = init_v1_2()
        this.Compact_InPlace(context, False)
        graph: LDGraph = this.Flatten()
        graph.SetContext(context)
        return ARCtrl_ROCrate_LDGraph__LDGraph_ToROCrateJsonString_71136F3F(graph, spaces)


Profile_reflection = _expr4128

def Profile__ctor_6B2DF90D(root_data_entity: RootDataEntity, license: License | None=None, ro_crate_spec: str | None=None) -> Profile:
    return Profile(root_data_entity, license, ro_crate_spec)


__all__ = ["CustomResourceDescriptor_reflection", "ResourceDescriptorType__get_ID", "ResourceDescriptorType__get_Role", "Organization_reflection", "Author_reflection", "UsedType_reflection", "License_reflection", "TextualResource_reflection", "ResourceDescriptor_reflection", "Specification_reflection", "Constraint_reflection", "Guidance_reflection", "Example_reflection", "RootDataEntity_reflection", "Profile_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.arc_types import ArcAssay
from ...arctrl_core.comment import Comment
from ...arctrl_core.data import DataAux_pathAndSelectorFromName
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.Helper.collections_ import (ResizeArray_distinct, ResizeArray_append, ResizeArray_choose, ResizeArray_collect, ResizeArray_filter, ResizeArray_map, ResizeArray_tryFind, Option_fromSeq, ResizeArray_groupBy, ResizeArray_singleton, ResizeArray_appendSingleton, Option_fromValueWithDefault)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_core.Table.arc_tables import ArcTables
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...arctrl_rocrate.LDTypes.file import LDFile
from ...arctrl_rocrate.LDTypes.lab_process import LDLabProcess
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...fable_library.list import of_seq
from ...fable_library.option import (default_arg, map)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from .basic import (BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_composePropertyValueFromOA_ZDED3A0F, BaseTypes_composeComment_Z13201A7E, BaseTypes_decomposeDefinedTerm_Z2F770004, BaseTypes_decomposePropertyValueToOA_Z2F770004, BaseTypes_decomposeComment_Z2F770004)
from .datamap import (DatamapConversion_composeFragmentDescriptors_Z8923FA3, DatamapConversion_decomposeFragmentDescriptors_Z6E59645F)
from .person import (PersonConversion_composePerson_Z64D846DC, PersonConversion_decomposePerson_Z6839B9E8)
from .table import (ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C, ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586)

def _expr3961() -> TypeInfo:
    return class_type("ARCtrl.Conversion.AssayConversion", None, AssayConversion)


class AssayConversion:
    ...

AssayConversion_reflection = _expr3961

def AssayConversion_getDataFilesFromProcesses_6BABD1B0(processes: Array[LDNode], fragment_descriptors: Array[LDNode] | None=None, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
    def f(df: LDNode, processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> LDNode | None:
        return LDPropertyValue.try_get_subject_of(df, graph, context)

    def f_1(p: LDNode, processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> Array[LDNode]:
        return ResizeArray_append(LDLabProcess.get_objects_as_data(p, graph, context), LDLabProcess.get_results_as_data(p, graph, context))

    data: Array[LDNode] = ResizeArray_distinct(ResizeArray_append(ResizeArray_choose(f, default_arg(fragment_descriptors, [])), ResizeArray_collect(f_1, processes)))
    def f_2(d: LDNode, processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> bool:
        return DataAux_pathAndSelectorFromName(d.Id)[1] is None

    files: Array[LDNode] = ResizeArray_filter(f_2, data)
    def f_7(tupled_arg: tuple[str, Array[LDNode]], processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> LDNode:
        path: str = tupled_arg[0]
        fragments: Array[LDNode] = tupled_arg[1]
        file: LDNode
        def f_5(d_3: LDNode, tupled_arg: Any=tupled_arg) -> bool:
            return d_3.Id == path

        match_value: LDNode | None = ResizeArray_tryFind(f_5, files)
        if match_value is None:
            comments: Array[LDNode] | None = Option_fromSeq(LDFile.get_comments(fragments[0], graph, context))
            file = LDFile.create(path, path, comments, LDFile.try_get_disambiguating_description_as_string(fragments[0], context), LDFile.try_get_encoding_format_as_string(fragments[0], context), None, fragments[0].TryGetContext())

        else: 
            file = match_value

        LDDataset.set_has_parts(file, fragments, context)
        return file

    def f_4(d_2: LDNode, processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> str:
        return DataAux_pathAndSelectorFromName(d_2.Id)[0]

    def f_3(d_1: LDNode, processes: Any=processes, fragment_descriptors: Any=fragment_descriptors, graph: Any=graph, context: Any=context) -> bool:
        return DataAux_pathAndSelectorFromName(d_1.Id)[1] is not None

    return ResizeArray_append(files, ResizeArray_map(f_7, ResizeArray_groupBy(f_4, ResizeArray_filter(f_3, data))))


def AssayConversion_composeAssay_Z5C53FD5C(assay: ArcAssay, fs: FileSystem | None=None) -> LDNode:
    def mapping(term: OntologyAnnotation, assay: Any=assay, fs: Any=fs) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term)

    measurement_method: LDNode | None = map(mapping, assay.TechnologyType)
    def mapping_1(term_1: OntologyAnnotation, assay: Any=assay, fs: Any=fs) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term_1)

    measurement_technique: LDNode | None = map(mapping_1, assay.TechnologyPlatform)
    def mapping_2(term_2: OntologyAnnotation, assay: Any=assay, fs: Any=fs) -> LDNode:
        return BaseTypes_composePropertyValueFromOA_ZDED3A0F(term_2)

    variable_measured: LDNode | None = map(mapping_2, assay.MeasurementType)
    def f(c: Person, assay: Any=assay, fs: Any=fs) -> LDNode:
        return PersonConversion_composePerson_Z64D846DC(c)

    creators: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, assay.Performers))
    process_sequence: Array[LDNode] | None = Option_fromSeq(list(ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C(ArcTables(assay.Tables), assay.Identifier, None, fs)))
    def mapping_3(datamap: DataMap, assay: Any=assay, fs: Any=fs) -> Array[LDNode]:
        return DatamapConversion_composeFragmentDescriptors_Z8923FA3(datamap)

    fragment_descriptors: Array[LDNode] | None = map(mapping_3, assay.DataMap)
    def mapping_4(ps: Array[LDNode], assay: Any=assay, fs: Any=fs) -> Array[LDNode]:
        return AssayConversion_getDataFilesFromProcesses_6BABD1B0(ps, fragment_descriptors)

    data_files: Array[LDNode] | None = map(mapping_4, process_sequence)
    def _arrow3969(__unit: None=None, assay: Any=assay, fs: Any=fs) -> Array[LDNode] | None:
        fds_1: Array[LDNode] = fragment_descriptors
        return fds_1

    def _arrow3970(__unit: None=None, assay: Any=assay, fs: Any=fs) -> Array[LDNode] | None:
        vm_1: LDNode = variable_measured
        return ResizeArray_singleton(vm_1)

    def _arrow3971(__unit: None=None, assay: Any=assay, fs: Any=fs) -> Array[LDNode] | None:
        fds: Array[LDNode] = fragment_descriptors
        vm: LDNode = variable_measured
        return ResizeArray_appendSingleton(vm, fds)

    variable_measureds: Array[LDNode] | None = (None if (fragment_descriptors is None) else _arrow3969()) if (variable_measured is None) else (_arrow3970() if (fragment_descriptors is None) else _arrow3971())
    def f_1(c_1: Comment, assay: Any=assay, fs: Any=fs) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(c_1)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_1, assay.Comments))
    return LDDataset.create_assay(assay.Identifier, None, assay.Title, assay.Description, creators, data_files, measurement_method, measurement_technique, variable_measureds, process_sequence, comments)


def AssayConversion_decomposeAssay_Z6839B9E8(assay: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcAssay:
    def mapping(m: LDNode, assay: Any=assay, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(m, context)

    measurement_method: OntologyAnnotation | None = map(mapping, LDDataset.try_get_measurement_method_as_defined_term(assay, graph, context))
    def mapping_1(m_1: LDNode, assay: Any=assay, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(m_1, context)

    measurement_technique: OntologyAnnotation | None = map(mapping_1, LDDataset.try_get_measurement_technique_as_defined_term(assay, graph, context))
    def mapping_2(v: LDNode, assay: Any=assay, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposePropertyValueToOA_Z2F770004(v, context)

    variable_measured: OntologyAnnotation | None = map(mapping_2, LDDataset.try_get_variable_measured_as_measurement_type(assay, graph, context))
    def f(c: LDNode, assay: Any=assay, graph: Any=graph, context: Any=context) -> Person:
        return PersonConversion_decomposePerson_Z6839B9E8(c, graph, context)

    perfomers: Array[Person] = ResizeArray_map(f, LDDataset.get_creators(assay, graph, context))
    data_map: DataMap | None
    v_1: DataMap = DatamapConversion_decomposeFragmentDescriptors_Z6E59645F(LDDataset.get_variable_measured_as_fragment_descriptors(assay, graph, context), graph, context)
    data_map = Option_fromValueWithDefault(DataMap.init(), v_1)
    tables: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586(of_seq(LDDataset.get_abouts_as_lab_process(assay, graph, context)), graph, context)
    def f_1(c_1: LDNode, assay: Any=assay, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c_1, context)

    comments: Array[Comment] = ResizeArray_map(f_1, LDDataset.get_comments(assay, graph, context))
    return ArcAssay.create(LDDataset.get_identifier_as_string(assay, context), LDDataset.try_get_name_as_string(assay, context), LDDataset.try_get_description_as_string(assay, context), variable_measured, measurement_method, measurement_technique, tables.Tables, data_map, perfomers, comments)


__all__ = ["AssayConversion_reflection", "AssayConversion_getDataFilesFromProcesses_6BABD1B0", "AssayConversion_composeAssay_Z5C53FD5C", "AssayConversion_decomposeAssay_Z6839B9E8"]




from __future__ import annotations
from ..arctrl_core.arc_types import (ArcAssay, ArcStudy, ArcWorkflow, ArcRun, ArcInvestigation)
from ..arctrl_file_system.file_system import FileSystem
from ..arctrl_rocrate.ldcontext import LDContext
from ..arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ..fable_library.reflection import (TypeInfo, class_type)
from .Conversion.assay import (AssayConversion_composeAssay_Z5C53FD5C, AssayConversion_decomposeAssay_Z6839B9E8)
from .Conversion.investigation import (InvestigationConversion_composeInvestigation_5AEC717D, InvestigationConversion_decomposeInvestigation_Z6839B9E8)
from .Conversion.run import (RunConversion_composeRun_Z8CC08AC, RunConversion_decomposeRun_Z6839B9E8)
from .Conversion.study import (StudyConversion_composeStudy_ZFE0E38E, StudyConversion_decomposeStudy_Z6839B9E8)
from .Conversion.workflow import (WorkflowConversion_composeWorkflow_42450E6E, WorkflowConversion_decomposeWorkflow_Z6839B9E8)

def ARCtrl_ArcAssay__ArcAssay_ToROCrateAssay_1695DD5C(this: ArcAssay, fs: FileSystem | None=None) -> LDNode:
    return AssayConversion_composeAssay_Z5C53FD5C(this, fs)


def ARCtrl_ArcAssay__ArcAssay_fromROCrateAssay_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcAssay:
    return AssayConversion_decomposeAssay_Z6839B9E8(a, graph, context)


def ARCtrl_ArcStudy__ArcStudy_ToROCrateStudy_1695DD5C(this: ArcStudy, fs: FileSystem | None=None) -> LDNode:
    return StudyConversion_composeStudy_ZFE0E38E(this, fs)


def ARCtrl_ArcStudy__ArcStudy_fromROCrateStudy_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcStudy:
    return StudyConversion_decomposeStudy_Z6839B9E8(a, graph, context)


def ARCtrl_ArcWorkflow__ArcWorkflow_ToROCrateWorkflow_1695DD5C(this: ArcWorkflow, fs: FileSystem | None=None) -> LDNode:
    return WorkflowConversion_composeWorkflow_42450E6E(this, fs)


def ARCtrl_ArcWorkflow__ArcWorkflow_fromROCrateWorkflow_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcWorkflow:
    return WorkflowConversion_decomposeWorkflow_Z6839B9E8(a, graph, context)


def ARCtrl_ArcRun__ArcRun_ToROCrateRun_1695DD5C(this: ArcRun, fs: FileSystem | None=None) -> LDNode:
    return RunConversion_composeRun_Z8CC08AC(this, fs)


def ARCtrl_ArcRun__ArcRun_fromROCrateRun_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcRun:
    return RunConversion_decomposeRun_Z6839B9E8(a, graph, context)


def ARCtrl_ArcInvestigation__ArcInvestigation_ToROCrateInvestigation_1695DD5C(this: ArcInvestigation, fs: FileSystem | None=None) -> LDNode:
    return InvestigationConversion_composeInvestigation_5AEC717D(this, fs)


def ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcInvestigation:
    return InvestigationConversion_decomposeInvestigation_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_toArcAssay_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcAssay:
    return AssayConversion_decomposeAssay_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_fromArcAssay_Static_1501C0F8(a: ArcAssay) -> LDNode:
    return AssayConversion_composeAssay_Z5C53FD5C(a)


def ARCtrl_ROCrate_Dataset__Dataset_toArcStudy_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcStudy:
    return StudyConversion_decomposeStudy_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_fromArcStudy_Static_1680536E(a: ArcStudy) -> LDNode:
    return StudyConversion_composeStudy_ZFE0E38E(a)


def ARCtrl_ROCrate_Dataset__Dataset_toArcWorkflow_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcWorkflow:
    return WorkflowConversion_decomposeWorkflow_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_fromArcWorkflow_Static_Z1C75CB0E(a: ArcWorkflow) -> LDNode:
    return WorkflowConversion_composeWorkflow_42450E6E(a)


def ARCtrl_ROCrate_Dataset__Dataset_toArcRun_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcRun:
    return RunConversion_decomposeRun_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_fromArcRun_Static_Z3EFAF6F8(a: ArcRun) -> LDNode:
    return RunConversion_composeRun_Z8CC08AC(a)


def ARCtrl_ROCrate_Dataset__Dataset_toArcInvestigation_Static_Z6839B9E8(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcInvestigation:
    return InvestigationConversion_decomposeInvestigation_Z6839B9E8(a, graph, context)


def ARCtrl_ROCrate_Dataset__Dataset_fromArcInvestigation_Static_Z720BD3FF(a: ArcInvestigation) -> LDNode:
    return InvestigationConversion_composeInvestigation_5AEC717D(a)


def _expr3992() -> TypeInfo:
    return class_type("ARCtrl.Conversion.TypeExtensions.Conversion", None, Conversion)


class Conversion:
    @staticmethod
    def arc_assay_to_dataset(a: ArcAssay, fs: FileSystem | None=None) -> LDNode:
        return ARCtrl_ArcAssay__ArcAssay_ToROCrateAssay_1695DD5C(a, fs)

    @staticmethod
    def dataset_to_arc_assay(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromROCrateAssay_Static_Z6839B9E8(a, graph, context)

    @staticmethod
    def arc_study_to_dataset(a: ArcStudy, fs: FileSystem | None=None) -> LDNode:
        return ARCtrl_ArcStudy__ArcStudy_ToROCrateStudy_1695DD5C(a, fs)

    @staticmethod
    def dataset_to_arc_study(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcStudy:
        return ARCtrl_ArcStudy__ArcStudy_fromROCrateStudy_Static_Z6839B9E8(a, graph, context)

    @staticmethod
    def arc_workflow_to_dataset(a: ArcWorkflow, fs: FileSystem | None=None) -> LDNode:
        return ARCtrl_ArcWorkflow__ArcWorkflow_ToROCrateWorkflow_1695DD5C(a, fs)

    @staticmethod
    def dataset_to_arc_workflow(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcWorkflow:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromROCrateWorkflow_Static_Z6839B9E8(a, graph, context)

    @staticmethod
    def arc_run_to_dataset(a: ArcRun, fs: FileSystem | None=None) -> LDNode:
        return ARCtrl_ArcRun__ArcRun_ToROCrateRun_1695DD5C(a, fs)

    @staticmethod
    def dataset_to_arc_run(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcRun:
        return ARCtrl_ArcRun__ArcRun_fromROCrateRun_Static_Z6839B9E8(a, graph, context)

    @staticmethod
    def arc_investigation_to_dataset(a: ArcInvestigation, fs: FileSystem | None=None) -> LDNode:
        return ARCtrl_ArcInvestigation__ArcInvestigation_ToROCrateInvestigation_1695DD5C(a, fs)

    @staticmethod
    def dataset_to_arc_investigation(a: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8(a, graph, context)


Conversion_reflection = _expr3992

__all__ = ["ARCtrl_ArcAssay__ArcAssay_ToROCrateAssay_1695DD5C", "ARCtrl_ArcAssay__ArcAssay_fromROCrateAssay_Static_Z6839B9E8", "ARCtrl_ArcStudy__ArcStudy_ToROCrateStudy_1695DD5C", "ARCtrl_ArcStudy__ArcStudy_fromROCrateStudy_Static_Z6839B9E8", "ARCtrl_ArcWorkflow__ArcWorkflow_ToROCrateWorkflow_1695DD5C", "ARCtrl_ArcWorkflow__ArcWorkflow_fromROCrateWorkflow_Static_Z6839B9E8", "ARCtrl_ArcRun__ArcRun_ToROCrateRun_1695DD5C", "ARCtrl_ArcRun__ArcRun_fromROCrateRun_Static_Z6839B9E8", "ARCtrl_ArcInvestigation__ArcInvestigation_ToROCrateInvestigation_1695DD5C", "ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_toArcAssay_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_fromArcAssay_Static_1501C0F8", "ARCtrl_ROCrate_Dataset__Dataset_toArcStudy_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_fromArcStudy_Static_1680536E", "ARCtrl_ROCrate_Dataset__Dataset_toArcWorkflow_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_fromArcWorkflow_Static_Z1C75CB0E", "ARCtrl_ROCrate_Dataset__Dataset_toArcRun_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_fromArcRun_Static_Z3EFAF6F8", "ARCtrl_ROCrate_Dataset__Dataset_toArcInvestigation_Static_Z6839B9E8", "ARCtrl_ROCrate_Dataset__Dataset_fromArcInvestigation_Static_Z720BD3FF", "Conversion_reflection"]




from typing import Any
from ...arctrl_json.decode import Decode_datetime
from ...arctrl_json.encode import date_time
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string as to_string_1

def try_from_string(s: str) -> Any | None:
    try: 
        def _arrow3873(__unit: None=None) -> Any:
            match_value: FSharpResult_2[Any, str] = Decode_fromString(Decode_datetime, s)
            if match_value.tag == 1:
                raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

            else: 
                return match_value.fields[0]


        return _arrow3873()

    except Exception as match_value_1:
        return None



def to_string(d: Any) -> str:
    return to_string_1(0, date_time(d))


__all__ = ["try_from_string", "to_string"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.arc_types import ArcInvestigation
from ..arctrl_core.Helper.collections_ import ResizeArray_choose
from ..arctrl_file_system.file_system import FileSystem
from ..arctrl_json.ROCrate.ldgraph import (encoder, decoder)
from ..arctrl_json.ROCrate.ldnode import decoder as decoder_1
from ..arctrl_rocrate.ldcontext import LDContext
from ..arctrl_rocrate.ldobject import (LDNode, LDRef, LDGraph)
from ..arctrl_rocrate.LDTypes.creative_work import LDCreativeWork
from ..arctrl_rocrate.LDTypes.dataset import LDDataset
from ..arctrl_rocrate.LDTypes.file import LDFile
from ..arctrl_rocrate.rocrate_context import (init_v1_2, init_bioschemas_context, init_v1_1)
from ..fable_library.date import now
from ..fable_library.option import bind
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import exactly_one
from ..fable_library.types import Array
from ..thoth_json_core.decode import map
from ..thoth_json_core.types import (IEncodable, Decoder_1)
from .conversion import (ARCtrl_ArcInvestigation__ArcInvestigation_ToROCrateInvestigation_1695DD5C, ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8)
from .license import License

def _expr3993() -> TypeInfo:
    return class_type("ARCtrl.Json.ARC.ROCrate", None, ROCrate)


class ROCrate:
    ...

ROCrate_reflection = _expr3993

def ROCrate_get_metadataFileDescriptor(__unit: None=None) -> LDNode:
    node: LDNode = LDNode("ro-crate-metadata.json", ["http://schema.org/CreativeWork"])
    node.SetProperty("http://purl.org/dc/terms/conformsTo", LDRef("https://w3id.org/ro/crate/1.2"))
    node.SetProperty("http://schema.org/about", LDRef("./"))
    return node


def ROCrate_createLicenseNode_29619109(license: License | None=None) -> LDNode:
    if license is None:
        return LDCreativeWork.create("#LICENSE", None, None, None, None, None, None, None, None, "ALL RIGHTS RESERVED BY THE AUTHORS")

    else: 
        license_1: License = license
        text: str
        match_value: str = license_1.Type
        text = license_1.Content
        return LDCreativeWork.create(license_1.Path, None, None, None, None, None, None, None, None, text)



def ROCrate_getLicense_Z2F770004(license: LDNode, context: LDContext | None=None) -> License | None:
    text: str | None = LDCreativeWork.try_get_text_as_string(license, context)
    match_value: str = license.Id
    (pattern_matching_result, text_1, path, text_2, path_1) = (None, None, None, None, None)
    if match_value == "#LICENSE":
        if text is not None:
            if text == "ALL RIGHTS RESERVED BY THE AUTHORS":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1
                text_1 = text


        else: 
            pattern_matching_result = 0


    elif text is None:
        pattern_matching_result = 3
        path_1 = match_value

    else: 
        pattern_matching_result = 2
        path = match_value
        text_2 = text

    if pattern_matching_result == 0:
        return None

    elif pattern_matching_result == 1:
        return License("fulltext", text_1)

    elif pattern_matching_result == 2:
        return License("fulltext", text_2, path)

    elif pattern_matching_result == 3:
        return License("fulltext", "", path_1)



def ROCrate_encoder_1E8A3F74(isa: ArcInvestigation, license: License | None=None, fs: FileSystem | None=None) -> IEncodable:
    license_1: LDNode = ROCrate_createLicenseNode_29619109(license)
    isa_1: LDNode = ARCtrl_ArcInvestigation__ArcInvestigation_ToROCrateInvestigation_1695DD5C(isa, fs)
    LDDataset.set_sddate_published_as_date_time(isa_1, now())
    LDDataset.set_license_as_creative_work(isa_1, license_1)
    graph: LDGraph = isa_1.Flatten()
    context: LDContext = LDContext(None, [init_v1_2(), init_bioschemas_context()])
    graph.SetContext(context)
    graph.AddNode(ROCrate_get_metadataFileDescriptor())
    graph.Compact_InPlace()
    return encoder(graph)


def ROCrate_get_decoder(__unit: None=None) -> Decoder_1[tuple[ArcInvestigation, Array[str], License | None]]:
    def ctor(graph: LDGraph) -> tuple[ArcInvestigation, Array[str], License | None]:
        match_value: LDNode | None = graph.TryGetNode("./")
        if match_value is None:
            raise Exception("RO-Crate graph did not contain root data Entity")

        else: 
            node: LDNode = match_value
            def f(n: LDNode, graph: Any=graph) -> str | None:
                if (not n.HasType(LDDataset.schema_type(), graph.TryGetContext())) if ((not (n.Id.find("#") >= 0)) if LDFile.validate(n, graph.TryGetContext()) else False) else False:
                    return n.Id

                else: 
                    return None


            files: Array[str] = ResizeArray_choose(f, graph.Nodes)
            def binder(n_1: LDNode, graph: Any=graph) -> License | None:
                return ROCrate_getLicense_Z2F770004(n_1, graph.TryGetContext())

            license: License | None = bind(binder, LDDataset.try_get_license_as_creative_work(node, graph, graph.TryGetContext()))
            return (ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8(node, graph, graph.TryGetContext()), files, license)


    return map(ctor, decoder)


def ROCrate_get_decoderDeprecated(__unit: None=None) -> Decoder_1[ArcInvestigation]:
    def ctor(ldnode: LDNode) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateInvestigation_Static_Z6839B9E8(exactly_one(LDDataset.get_abouts(ldnode)), None, init_v1_1())

    return map(ctor, decoder_1)


__all__ = ["ROCrate_reflection", "ROCrate_get_metadataFileDescriptor", "ROCrate_createLicenseNode_29619109", "ROCrate_getLicense_Z2F770004", "ROCrate_encoder_1E8A3F74", "ROCrate_get_decoder", "ROCrate_get_decoderDeprecated"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_contract.contract import (Contract, DTOType, DTO as DTO_2)
from ...arctrl_file_system.path import combine
from ...fable_library.array_ import (fold, append)
from ...fable_library.async_builder import (singleton, Async)
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import Array
from ...fable_library.util import curry2
from ...fs_spreadsheet.fs_workbook import FsWorkbook
from ..cross_async import (catch_with, start_sequential)
from .file_system_helper import (read_file_xlsx_async, read_file_text_async, ensure_directory_of_file_async, write_file_text_async, write_file_xlsx_async, rename_file_or_directory_async, delete_file_or_directory_async)

def fulfill_read_contract_async(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e_1: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4033(__unit: None=None, e_1: Any=e_1) -> str:
            arg_4: str = str(e_1)
            return to_text(printf("Error reading contract %s: %s"))(c.Path)(arg_4)

        return FSharpResult_2(1, _arrow4033())

    def _arrow4039(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        def _arrow4036(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
            match_value: DTOType | None = c.DTOType
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value.tag == 0:
                    pattern_matching_result = 0

                elif match_value.tag == 4:
                    pattern_matching_result = 0

                elif match_value.tag == 1:
                    pattern_matching_result = 0

                elif match_value.tag == 2:
                    pattern_matching_result = 0

                elif match_value.tag == 3:
                    pattern_matching_result = 0

                elif match_value.tag == 5:
                    pattern_matching_result = 0

                elif match_value.tag == 10:
                    pattern_matching_result = 1

                elif match_value.tag == 8:
                    pattern_matching_result = 1

                elif match_value.tag == 9:
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                path: str = combine(base_path, c.Path)
                def _arrow4034(_arg: FsWorkbook) -> Async[FSharpResult_2[Contract, str]]:
                    return singleton.Return(FSharpResult_2(0, Contract(c.Operation, c.Path, c.DTOType, DTO_2(0, _arg))))

                return singleton.Bind(read_file_xlsx_async(path), _arrow4034)

            elif pattern_matching_result == 1:
                path_1: str = combine(base_path, c.Path)
                def _arrow4035(_arg_1: str) -> Async[FSharpResult_2[Contract, str]]:
                    return singleton.Return(FSharpResult_2(0, Contract(c.Operation, c.Path, c.DTOType, DTO_2(1, _arg_1))))

                return singleton.Bind(read_file_text_async(path_1), _arrow4035)

            elif pattern_matching_result == 2:
                return singleton.Return(FSharpResult_2(1, to_text(printf("Contract %s is neither an ISA nor a freetext contract"))(c.Path)))


        def _arrow4038(_arg_2: Exception) -> Async[FSharpResult_2[Contract, str]]:
            def _arrow4037(__unit: None=None) -> str:
                arg_2: str = str(_arg_2)
                return to_text(printf("Error reading contract %s: %s"))(c.Path)(arg_2)

            return singleton.Return(FSharpResult_2(1, _arrow4037()))

        return singleton.TryWith(singleton.Delay(_arrow4036), _arrow4038)

    return catch_with(f, singleton.Delay(_arrow4039))


def fullfill_contract_batch_async_by(contract_f: Callable[[str, Contract], Async[FSharpResult_2[Contract, str]]], base_path: str, cs: Array[Contract]) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
    def _arrow4041(__unit: None=None, contract_f: Any=contract_f, base_path: Any=base_path, cs: Any=cs) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        def _arrow4040(_arg: Array[FSharpResult_2[Contract, str]]) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
            def folder(acc: FSharpResult_2[Array[Contract], Array[str]], cr: FSharpResult_2[Contract, str]) -> FSharpResult_2[Array[Contract], Array[str]]:
                copy_of_struct: FSharpResult_2[Array[Contract], Array[str]] = acc
                if copy_of_struct.tag == 1:
                    copy_of_struct_1: FSharpResult_2[Contract, str] = cr
                    if copy_of_struct_1.tag == 1:
                        return FSharpResult_2(1, append(copy_of_struct.fields[0], [copy_of_struct_1.fields[0]], None))

                    else: 
                        return FSharpResult_2(1, copy_of_struct.fields[0])


                else: 
                    copy_of_struct_2: FSharpResult_2[Contract, str] = cr
                    if copy_of_struct_2.tag == 1:
                        return FSharpResult_2(1, [copy_of_struct_2.fields[0]])

                    else: 
                        return FSharpResult_2(0, append(copy_of_struct.fields[0], [copy_of_struct_2.fields[0]], None))



            res: FSharpResult_2[Array[Contract], Array[str]] = fold(folder, FSharpResult_2(0, []), _arg)
            return singleton.Return(res)

        return singleton.Bind(start_sequential(curry2(contract_f)(base_path), cs), _arrow4040)

    return singleton.Delay(_arrow4041)


def fulfill_write_contract_async(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e_1: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4042(__unit: None=None, e_1: Any=e_1) -> str:
            arg_4: str = str(e_1)
            return to_text(printf("Error writing contract %s: %s"))(c.Path)(arg_4)

        return FSharpResult_2(1, _arrow4042())

    def _arrow4052(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        def _arrow4049(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
            match_value: DTO_2 | None = c.DTO
            if match_value is None:
                path_2: str = combine(base_path, c.Path)
                def _arrow4044(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4043(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_text_async(path_2, ""), _arrow4043)

                return singleton.Bind(ensure_directory_of_file_async(path_2), _arrow4044)

            elif match_value.tag == 1:
                t: str = match_value.fields[0]
                path: str = combine(base_path, c.Path)
                def _arrow4046(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4045(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_text_async(path, t), _arrow4045)

                return singleton.Bind(ensure_directory_of_file_async(path), _arrow4046)

            elif match_value.tag == 0:
                wb: Any = match_value.fields[0]
                path_1: str = combine(base_path, c.Path)
                def _arrow4048(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4047(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_xlsx_async(path_1, wb), _arrow4047)

                return singleton.Bind(ensure_directory_of_file_async(path_1), _arrow4048)

            else: 
                return singleton.Return(FSharpResult_2(1, to_text(printf("Contract %s is not an ISA contract"))(c.Path)))


        def _arrow4051(_arg_6: Exception) -> Async[FSharpResult_2[Contract, str]]:
            def _arrow4050(__unit: None=None) -> str:
                arg_2: str = str(_arg_6)
                return to_text(printf("Error writing contract %s: %s"))(c.Path)(arg_2)

            return singleton.Return(FSharpResult_2(1, _arrow4050()))

        return singleton.TryWith(singleton.Delay(_arrow4049), _arrow4051)

    return catch_with(f, singleton.Delay(_arrow4052))


def fulfill_update_contract_async(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e_1: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4053(__unit: None=None, e_1: Any=e_1) -> str:
            arg_4: str = str(e_1)
            return to_text(printf("Error updating contract %s: %s"))(c.Path)(arg_4)

        return FSharpResult_2(1, _arrow4053())

    def _arrow4063(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        def _arrow4060(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
            match_value: DTO_2 | None = c.DTO
            if match_value is None:
                path_2: str = combine(base_path, c.Path)
                def _arrow4055(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4054(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_text_async(path_2, ""), _arrow4054)

                return singleton.Bind(ensure_directory_of_file_async(path_2), _arrow4055)

            elif match_value.tag == 1:
                t: str = match_value.fields[0]
                path: str = combine(base_path, c.Path)
                def _arrow4057(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4056(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_text_async(path, t), _arrow4056)

                return singleton.Bind(ensure_directory_of_file_async(path), _arrow4057)

            elif match_value.tag == 0:
                wb: Any = match_value.fields[0]
                path_1: str = combine(base_path, c.Path)
                def _arrow4059(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    def _arrow4058(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                        return singleton.Return(FSharpResult_2(0, c))

                    return singleton.Bind(write_file_xlsx_async(path_1, wb), _arrow4058)

                return singleton.Bind(ensure_directory_of_file_async(path_1), _arrow4059)

            else: 
                return singleton.Return(FSharpResult_2(1, to_text(printf("Contract %s is not an ISA contract"))(c.Path)))


        def _arrow4062(_arg_6: Exception) -> Async[FSharpResult_2[Contract, str]]:
            def _arrow4061(__unit: None=None) -> str:
                arg_2: str = str(_arg_6)
                return to_text(printf("Error updating contract %s: %s"))(c.Path)(arg_2)

            return singleton.Return(FSharpResult_2(1, _arrow4061()))

        return singleton.TryWith(singleton.Delay(_arrow4060), _arrow4062)

    return catch_with(f, singleton.Delay(_arrow4063))


def fullfill_rename_contract_async(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e_1: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4064(__unit: None=None, e_1: Any=e_1) -> str:
            arg_5: str = str(e_1)
            return to_text(printf("Error renaming contract %s: %s"))(c.Path)(arg_5)

        return FSharpResult_2(1, _arrow4064())

    def _arrow4069(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        def _arrow4066(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
            match_value: DTO_2 | None = c.DTO
            (pattern_matching_result, t_2) = (None, None)
            if match_value is not None:
                if match_value.tag == 1:
                    if match_value.fields[0] == c.Path:
                        pattern_matching_result = 0

                    else: 
                        pattern_matching_result = 1
                        t_2 = match_value.fields[0]


                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                return singleton.Return(FSharpResult_2(1, to_text(printf("Rename Contract %s old and new Path are the same"))(c.Path)))

            elif pattern_matching_result == 1:
                new_path: str = combine(base_path, t_2)
                old_path: str = combine(base_path, c.Path)
                def _arrow4065(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                    return singleton.Return(FSharpResult_2(0, c))

                return singleton.Bind(rename_file_or_directory_async(old_path, new_path), _arrow4065)

            elif pattern_matching_result == 2:
                return singleton.Return(FSharpResult_2(1, to_text(printf("Rename Contract %s does not contain new Path"))(c.Path)))


        def _arrow4068(_arg_1: Exception) -> Async[FSharpResult_2[Contract, str]]:
            def _arrow4067(__unit: None=None) -> str:
                arg_3: str = str(_arg_1)
                return to_text(printf("Error renaming contract %s: %s"))(c.Path)(arg_3)

            return singleton.Return(FSharpResult_2(1, _arrow4067()))

        return singleton.TryWith(singleton.Delay(_arrow4066), _arrow4068)

    return catch_with(f, singleton.Delay(_arrow4069))


def fullfill_delete_contract_async(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e_1: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4070(__unit: None=None, e_1: Any=e_1) -> str:
            arg_3: str = str(e_1)
            return to_text(printf("Error deleting contract %s: %s"))(c.Path)(arg_3)

        return FSharpResult_2(1, _arrow4070())

    def _arrow4077(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        def _arrow4074(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
            path: str = combine(base_path, c.Path)
            def _arrow4073(__unit: None=None) -> Async[FSharpResult_2[Contract, str]]:
                return singleton.Return(FSharpResult_2(0, c))

            return singleton.Bind(delete_file_or_directory_async(path), _arrow4073)

        def _arrow4076(_arg_1: Exception) -> Async[FSharpResult_2[Contract, str]]:
            def _arrow4075(__unit: None=None) -> str:
                arg_1: str = str(_arg_1)
                return to_text(printf("Error deleting contract %s: %s"))(c.Path)(arg_1)

            return singleton.Return(FSharpResult_2(1, _arrow4075()))

        return singleton.TryWith(singleton.Delay(_arrow4074), _arrow4076)

    return catch_with(f, singleton.Delay(_arrow4077))


def full_fill_contract(base_path: str, c: Contract) -> Async[FSharpResult_2[Contract, str]]:
    def f(e: Exception, base_path: Any=base_path, c: Any=c) -> FSharpResult_2[Contract, str]:
        def _arrow4078(__unit: None=None, e: Any=e) -> str:
            arg_2: str = str(e)
            return to_text(printf("Error fulfilling contract %s: %s"))(c.Path)(arg_2)

        return FSharpResult_2(1, _arrow4078())

    def _arrow4079(__unit: None=None, base_path: Any=base_path, c: Any=c) -> Async[FSharpResult_2[Contract, str]]:
        match_value: str = c.Operation
        return singleton.ReturnFrom(fulfill_read_contract_async(base_path, c)) if (match_value == "READ") else (singleton.ReturnFrom(fulfill_write_contract_async(base_path, c)) if (match_value == "CREATE") else (singleton.ReturnFrom(fulfill_update_contract_async(base_path, c)) if (match_value == "UPDATE") else (singleton.ReturnFrom(fullfill_delete_contract_async(base_path, c)) if (match_value == "DELETE") else (singleton.ReturnFrom(fullfill_rename_contract_async(base_path, c)) if (match_value == "RENAME") else singleton.Return(FSharpResult_2(1, to_text(printf("Operation %A not supported"))(c.Operation)))))))

    return catch_with(f, singleton.Delay(_arrow4079))


def full_fill_contract_batch_async(base_path: str, cs: Array[Contract]) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
    def _arrow4080(base_path_1: str, c: Contract, base_path: Any=base_path, cs: Any=cs) -> Async[FSharpResult_2[Contract, str]]:
        return full_fill_contract(base_path_1, c)

    return fullfill_contract_batch_async_by(_arrow4080, base_path, cs)


__all__ = ["fulfill_read_contract_async", "fullfill_contract_batch_async_by", "fulfill_write_contract_async", "fulfill_update_contract_async", "fullfill_rename_contract_async", "fullfill_delete_contract_async", "full_fill_contract", "full_fill_contract_batch_async"]




from typing import Any
from ..fable_library.reflection import (TypeInfo, union_type)
from ..fable_library.types import (Array, Union)
from .tool_description import CWLToolDescription_reflection
from .workflow_description import CWLWorkflowDescription_reflection

def _expr518() -> TypeInfo:
    return union_type("ARCtrl.CWL.CWLProcessingUnit", [], CWLProcessingUnit, lambda: [[("Item", CWLToolDescription_reflection())], [("Item", CWLWorkflowDescription_reflection())]])


class CWLProcessingUnit(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["CommandLineTool", "Workflow"]


CWLProcessingUnit_reflection = _expr518

__all__ = ["CWLProcessingUnit_reflection"]

